import os
import platform
import time
from argparse import Namespace
from pathlib import Path
from subprocess import Popen, STDOUT
from typing import Optional, List, Tuple, Generator, Union
from zipfile import ZipFile

import psutil
import requests
import toml
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TransferSpeedColumn

from trecover.config import var, log


def download(direct_link: str, filepath: Path) -> None:
    """
    Download file.

    Parameters
    ----------
    direct_link : str
        Sharing link to the file on GutHub.
    filepath : Path
        Path to the downloaded file.

   """

    filepath.parent.mkdir(parents=True, exist_ok=True)

    with filepath.open(mode='wb') as fw:
        response = requests.get(direct_link, stream=True)
        total_length = response.headers.get('content-length')

        if total_length is None:  # no content length header
            fw.write(response.content)
        else:
            data_stream = response.iter_content(chunk_size=4096)

            with Progress(
                    TextColumn('{task.description}', style='bright_blue'),
                    BarColumn(complete_style='bright_blue'),
                    DownloadColumn(),
                    TransferSpeedColumn(),
                    transient=True,
                    console=log.project_console
            ) as progress:
                download_progress = progress.add_task('Downloading', total=int(total_length))

                for data in data_stream:
                    fw.write(data)
                    progress.update(download_progress, advance=4096)


def get_real_direct_link(sharing_link: str) -> str:
    """
    Get a direct download link.

    Parameters
    ----------
    sharing_link : str
        Sharing link to the file on Yandex disk.

    Returns
    -------
    str:
        Direct link if it converts, otherwise None.

    """

    pk_request = requests.get(
        f'https://cloud-api.yandex.net/v1/disk/public/resources/download?public_key={sharing_link}')

    return pk_request.json().get('href')


def extract_filename(direct_link: str) -> Optional[str]:
    """ Get filename of downloaded data """

    for chunk in direct_link.strip().split('&'):
        if chunk.startswith('filename='):
            return chunk.split('=')[1]

    return None


def download_from_disk(sharing_link: str, save_dir: Path) -> Optional[Path]:
    """
    Download file from Yandex disk.

    Parameters
    ----------
    sharing_link : str
        Sharing link to the file on Yandex disk.
    save_dir : Path
        Path where to store downloaded file.

    Returns
    -------
    filepath : Optional[Path]
        Path to the file if download was successful, otherwise None.

    """

    if not (direct_link := get_real_direct_link(sharing_link)):
        log.project_logger.error(f'[red]Failed to download data from [/][bright_blue] {sharing_link}')
        return None

    filename = extract_filename(direct_link) or 'downloaded_data'  # Try to recover the filename from the link
    filepath = save_dir / filename

    download(direct_link=direct_link, filepath=filepath)

    log.project_console.print(f'Downloaded "{filename}" to {filepath.absolute()}', style='green')

    return filepath


def download_from_github(direct_link: str, save_dir: Path) -> Path:
    """
    Download file from GutHub assets.

    Parameters
    ----------
    direct_link : str
        Sharing link to the file on GutHub.
    save_dir : Path
        Path where to store downloaded file.

    Returns
    -------
    filepath : Path
        Path to the downloaded file.

   """

    filename = direct_link.split('/')[-1]
    filepath = save_dir / filename

    download(direct_link=direct_link, filepath=filepath)

    log.project_console.print(f'Downloaded "{filename}" to {filepath.absolute()}', style='green')

    return filepath


def download_archive(link: str, save_dir: Path, yandex_disk: bool = False) -> None:
    """
    Download archive file and extract it to save_dir.

    Parameters
    ----------
    link : str
        Sharing link to the archive file on Yandex disk or GitHub assets.
    save_dir : Path
        Path where to store extracted data
    yandex_disk : bool, default=False
        If the link is to Yandex disk.

    """

    filepath = download_from_disk(link, save_dir) if yandex_disk else download_from_github(link, save_dir)

    if filepath:
        with ZipFile(filepath) as zf:
            zf.extractall(path=save_dir)

        os.remove(filepath)

        log.project_console.print(f'Archive extracted to {save_dir.absolute()}', style='green')


def get_files_columns(inference_path: Path,
                      separator: str,
                      noisy: bool,
                      min_noise: int,
                      max_noise: int,
                      n_to_show: int,
                      ) -> Tuple[List[Path], List[List[str]]]:
    """
    Get columns for keyless reading from plain data contained in the files with defined noise range.

    Parameters
    ----------
    inference_path : Path
        Paths to folder with files that contain data to read or create noised columns for keyless reading.
    separator : str
        Separator to split the data into columns.
    noisy : bool
        Indicates that the data in the files is already noisy and contains columns for keyless reading.
    min_noise : int
        Minimum noise range value.
    max_noise : int
        Maximum noise range value.
    n_to_show : int
        Maximum number of columns. Zero means no restrictions.

    Returns
    -------
    (files, files_columns) : Tuple[List[Path], List[List[str]]]
        List of paths and batch of columns for keyless reading.

    """

    from trecover.utils.inference import read_files_columns, create_files_noisy_columns

    if inference_path.is_file():
        files = [inference_path, ]
    else:
        files = [file for file in inference_path.iterdir()]

    if noisy:
        files_columns = read_files_columns(files, separator, n_to_show)
    else:
        files_columns = create_files_noisy_columns(files, min_noise, max_noise, n_to_show)

    return files, files_columns


def parse_config(file: Path) -> Namespace:
    """ Parse configuration file for 'trecover up' command. """

    conf = var.DEFAULT_CONFIG
    parsed_conf = toml.load(file)

    for service, params in parsed_conf.items():
        for variable, value in params.items():
            conf[service][variable] = value

    for service, params in conf.items():
        conf[service] = Namespace(**params)

    return Namespace(**conf)


def start_service(argv: List[str], name: str, logfile: Path, pidfile: Path) -> None:
    """
    Start service as a new process with given pid and log files.

    Parameters
    ----------
    argv : List[str]
        New process command.
    name : str
        Service name.
    logfile : Path
        Service logfile path.
    pidfile : Path
        Service pidfile path.

    """

    if platform.system() == 'Windows':
        from subprocess import CREATE_NO_WINDOW

        process = Popen(argv, creationflags=CREATE_NO_WINDOW, stdout=logfile.open(mode='w+'), stderr=STDOUT,
                        universal_newlines=True, start_new_session=True)
    else:
        process = Popen(argv, stdout=logfile.open(mode='w+'), stderr=STDOUT, universal_newlines=True,
                        start_new_session=True)

    with pidfile.open('w') as f:
        f.write(str(process.pid))

    log.project_console.print(f'The {name} service is started', style='bright_blue')


def stop_service(name: str, pidfile: Path, logfile: Path) -> None:
    """
    Send an interrupt signal to the process with given pid.

    Parameters
    ----------
    name : str
        Service name.
    pidfile : Path
        Service pidfile path.
    logfile : Path
        Service logfile path.

    """

    try:
        with pidfile.open() as f:
            pid = int(f.read())

        service = psutil.Process(pid=pid)

        for child_proc in service.children(recursive=True):
            child_proc.kill()

        if platform.system() != 'Windows':
            service.kill()

        service.wait()

        if logfile.exists():
            os.remove(logfile)

    except ValueError:
        log.project_console.print(f'The {name} service could not be stopped correctly'
                                  ' because its PID file is corrupted', style='red')
    except (OSError, psutil.NoSuchProcess):
        log.project_console.print(f'The {name} service could not be stopped correctly'
                                  ' because it probably failed earlier', style='red')
    else:
        log.project_console.print(f'The {name} service is stopped', style='bright_blue')

    finally:
        if pidfile.exists():
            os.remove(pidfile)


def check_service(name: str, pidfile: Path) -> None:
    """
    Display status of the process with given pid.

    Parameters
    ----------
    name : str
        Service name.
    pidfile : Path
        Service pidfile path.

    """

    try:
        with pidfile.open() as f:
            pid = int(f.read())

        if psutil.pid_exists(pid):
            log.project_console.print(f':rocket: The {name} status: running', style='bright_blue')
        else:
            log.project_console.print(f'The {name} status: dead', style='red')

    except FileNotFoundError:
        log.project_console.print(f'The {name} service is not started', style='yellow')

    except ValueError:
        log.project_console.print(f'The {name} service could not be checked correctly'
                                  ' because its PID file is corrupted', style='red')


def stream(*services: Union[Tuple[str, Path], Tuple[Tuple[str, Path]]],
           live: bool = False,
           period: float = 0.1
           ) -> Optional[Generator[str, None, None]]:
    """
    Get a generator that yields the services' stdout streams.

    Parameters
    ----------
    *services : Union[Tuple[str, Path], Tuple[Tuple[str, Path]]]
        Sequence of services' names and logfile's paths.
    live : bool, default=False
        Yield only new services' logs.
    period : float, default=0.1
        Generator's delay.

    Returns
    -------
    Optional[Generator[str, None, None]]:
        Generator that yields the services' stdout streams or None if services are stopped.

    """

    names = list()
    streams = list()
    alignment = 0

    try:
        for (name, log_file) in services:
            if not log_file.exists():
                continue

            names.append(name)

            service_stream = log_file.open()

            if live:
                service_stream.seek(0, 2)

            streams.append(service_stream)

            if len(name) > alignment:
                alignment = len(name)

        if not (n_services := len(names)):
            return None

        while True:
            for i in range(n_services):
                if record := streams[i].read().strip():
                    color = var.COLORS[i % len(var.COLORS)]

                    for record_line in record.split('\n'):
                        yield f'[{color}]{names[i]: <{alignment}} |[/{color}] {record_line}'

            time.sleep(period)

    finally:
        for service_stream in streams:
            service_stream.close()



import platform
from pathlib import Path
from typing import List, Tuple, Optional

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader

from trecover.config import var, log


class BaseCollate(object):
    # TODO docs

    def __init__(self, device: Optional[torch.device] = None):
        self.device = device or torch.device("cpu")

    def __str__(self) -> str:
        raise NotImplementedError

    def __call__(self, batch: List[str]) -> Tuple[Tensor, Tensor, Tensor, Optional[Tensor], Optional[Tensor], Tensor]:
        raise NotImplementedError

    def generate_subsequent_mask(self, size: int) -> Tensor:
        # TODO docs

        return torch.triu(torch.ones((size, size), dtype=torch.float, device=self.device), diagonal=1) == 1


class StandardCollate(BaseCollate):
    # TODO docs

    def __init__(self, min_noise: int, max_noise: int, device: Optional[torch.device] = None):
        # TODO docs

        assert 0 <= min_noise <= len(var.ALPHABET), \
            f'min_noise should be between 0 and {len(var.ALPHABET)} inclusive'
        assert min_noise <= max_noise <= len(var.ALPHABET), \
            f'max_noise should be between {min_noise} and {len(var.ALPHABET)} inclusive'

        super().__init__(device)
        self.min_noise = min_noise
        self.max_noise = max_noise

    def __str__(self) -> str:
        return f'<Collate(min_noise={self.min_noise}, max_noise={self.max_noise})>'

    def __call__(self, batch: List[str]) -> Tuple[Tensor, Tensor, Tensor, Optional[Tensor], Optional[Tensor], Tensor]:
        # TODO docs

        batch = [list(entry) for entry in batch]
        sizes = [len(entry) for entry in batch]
        batch_size, seq_len, token_size = len(batch), max(sizes), len(var.ALPHABET)

        src = torch.zeros((batch_size, seq_len, token_size), dtype=torch.float, device=self.device)
        tgt_inp = torch.zeros((batch_size, seq_len, token_size), dtype=torch.float, device=self.device)
        tgt = list()

        for i in range(len(batch)):

            i_tgt = torch.full((seq_len,), fill_value=-1, dtype=torch.long, device=self.device)

            for j in range(len(batch[i])):
                num_repr = var.ALPHABET2NUM[batch[i][j]]

                src[i, j, num_repr] = 1
                tgt_inp[i, j, num_repr] = 1
                i_tgt[j] = num_repr

                noise_size = np.random.randint(low=self.min_noise, high=self.max_noise + 1, size=1)[0]
                noise_indexes = np.random.randint(low=0, high=len(var.ALPHABET), size=noise_size)

                src[i, j, noise_indexes] = 1

            tgt.append(i_tgt)

        empty_token = torch.zeros((batch_size, 1, token_size), device=self.device)
        tgt_inp = torch.cat([empty_token, tgt_inp[:, :-1, :]], dim=1)
        tgt = torch.stack(tgt)
        subsequent_mask = self.generate_subsequent_mask(seq_len)

        if min(sizes) == seq_len:
            src_pad_mask = None
            tgt_inp_pad_mask = None

        else:
            src_pad_mask = torch.zeros((batch_size, seq_len), dtype=torch.bool, device=self.device)

            for i in range(len(batch)):
                src_pad_mask[i, sizes[i]:] = True

            empty_token_pad_mask = torch.zeros((batch_size, 1), dtype=torch.bool, device=self.device)
            tgt_inp_pad_mask = torch.cat([empty_token_pad_mask, src_pad_mask[:, :-1]], dim=1)

        return src, tgt_inp, tgt, src_pad_mask, tgt_inp_pad_mask, subsequent_mask


class WikiDataset(Dataset):
    # TODO docs

    def __init__(self, datafiles: List[Path], min_threshold: int, max_threshold: int, dataset_size: int):
        # TODO docs
        assert self.__exists(datafiles), 'datafiles do not exist'
        assert min_threshold > 0, 'min_threshold should be grater than 0'
        assert max_threshold >= min_threshold, f'max_threshold should be grater or equal than {min_threshold}'
        assert dataset_size > 0, 'dataset_size should be grater than 0'

        self.datafiles = datafiles
        self.n_files = range(len(self.datafiles))
        self.file_sizes = [file.stat().st_size for file in self.datafiles]
        self.distribution = self.__get_distribution()
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold
        self.dataset_size = dataset_size

    def __getitem__(self, idx: int) -> str:
        # TODO docs
        np.random.seed(None)

        file_id = np.random.choice(self.n_files, p=self.distribution)
        shift = np.random.randint(low=0, high=self.file_sizes[file_id] - self.max_threshold + 1, size=1)[0]
        line_size = np.random.randint(low=self.min_threshold, high=self.max_threshold + 1, size=1)[0]

        with open(self.datafiles[file_id], mode="r") as f:
            f.seek(shift)

            return f.read(line_size)

    def __len__(self) -> int:
        return self.dataset_size

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f'WikiDataset(min_threshold={self.min_threshold}, max_threshold={self.max_threshold},' \
               f' dataset_size={self.dataset_size})'

    @staticmethod
    def __exists(datafiles: List[Path]) -> bool:
        # TODO docs
        for file in datafiles:
            if not file.exists():
                print(f'{file} doesnt exist')
                return False

        return True if len(datafiles) else False

    def __get_distribution(self) -> np.ndarray:
        # TODO docs
        powered = np.array(self.file_sizes)

        return powered / np.sum(powered)

    def create_dataloader(self, batch_size: int,
                          collate: Optional[BaseCollate] = StandardCollate(min_noise=0, max_noise=0),
                          num_workers: int = 0,
                          pin_memory: bool = True
                          ) -> DataLoader:
        # TODO docs

        assert batch_size > 0, 'batch_size should be grater than 0'
        assert num_workers >= 0, 'num_workers should be grater or equal than 0'

        if collate.device.type == 'cuda':
            if platform.system() == 'Windows' and num_workers > 0:
                log.project_console.print('WARNING: Dataloader does not support num_workers > 0 and GPU device '
                                          'on Windows. The training data will be transferred to the GPU just '
                                          'before it is fed into the model, but not during batch generation.',
                                          style='bold red')
                collate.device = torch.device('cpu')

            pin_memory = False

        return DataLoader(dataset=self,
                          batch_size=batch_size,
                          shuffle=False,
                          num_workers=num_workers,
                          pin_memory=pin_memory,
                          collate_fn=collate)



from typer import Typer, Context, Option

cli = Typer(name='Train-cli', add_completion=False, help='Manage training')


@cli.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
             add_help_option=False,
             help='Start local training')
def local(ctx: Context,
          show_help: bool = Option(False, '--help', '-h', is_flag=True, help='Show help message and exit.')):
    """
    Start local training.

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.
    show_help : bool, default=False
        Show local train options.

    """

    from trecover.config import log
    from trecover.train.local import train, get_parser

    if show_help:
        get_parser().print_help()
    else:
        try:
            train(args=ctx.args)
        except Exception as e:
            log.project_logger.error(e)
            log.project_console.print_exception()
            log.error_console.print_exception(show_locals=True)


@cli.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
             add_help_option=False,
             help='Start remote training')
def remote(ctx: Context,
           show_help: bool = Option(False, '--help', '-h', is_flag=True, help='Show help message and exit.')
           ):
    """
    Start remote training.

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.
    show_help : bool, default=False
        Show remote train options.

    """

    pass


if __name__ == '__main__':
    cli()



import asyncio
from typing import Optional, List, Tuple, Callable, Awaitable

import celery
import torch
import torch.nn.functional as F
from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn, TimeElapsedColumn
from torch import Tensor

from trecover.config import log
from trecover.model import TRecover
from trecover.utils.transform import tensor_to_target
from trecover.utils.visualization import visualize_target


# ------------------------------------------Beam Search Auxiliary Functions---------------------------------------------


def get_steps_params(src: Tensor) -> Tuple[Tensor, List[int]]:
    """
    Get probability masks and beam widths required for each step of beam search algorithm.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.

    Returns
    -------
    steps_mask : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Probability masks that consists of zeros in the places that correspond to the letters allowed
        for selection in the column (src[i]) and values equal to minus infinity in all others.
    steps_width : List[int]
        The beam width for each step of the algorithm.

    """

    return (
        torch.full_like(src, fill_value=float('-inf')).masked_fill(src == 1, value=0.0),
        src.sum(dim=-1).int().tolist()
    )


# ----------------------------------------------Synchronous Beam Search-------------------------------------------------


def beam_step(candidates: List[Tuple[Tensor, float]],
              step_mask: Tensor,
              step_width: int,
              encoded_src: Tensor,
              model: TRecover,
              beam_width: int,
              device: torch.device
              ) -> List[Tuple[Tensor, float]]:
    """
    Implementation of the beam search algorithm step.

    Parameters
    ----------
    candidates : List[Tuple[Tensor[1, STEP_NUMBER, TOKEN_SIZE], float]]
        List of candidates from the previous step.
    step_mask : Tensor[TOKEN_SIZE]
        Column's mask that consists of zeros in the places that correspond to the letters allowed
        for selection in the column and values equal to minus infinity in all others.
        Required so that only the letters in the column are selected as a candidates.
    step_width : int
        Number of candidates that are contained in the step column.
    encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
        Columns for keyless reading that were encoded by TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    beam_width : int
        Number of candidates that can be selected at the current step.
    device : torch.device
        Device on which to allocate the candidate chains.

    Returns
    -------
    step_candidates : List[Tuple[Tensor[1, STEP_NUMBER + 1, TOKEN_SIZE], float]]
        List of candidates of size "beam_width" for the current step
        sorted in descending order of their probabilities.

    Notes
    -----
    For each chain candidate from the previous step:
    *       Probability distribution is calculated using trained TRecover model to
            select the next symbol from the current column,taking into account the "step_mask".
    *       The most probable symbols are selected from the calculated probability distribution,
            the number of which is set by the "step_width" and "beam_width" parameters.
    *       For each selected symbol, a new candidate chain with updated probability
            is constructed and placed in the "step_candidates" list.

    All candidates are sorted in descending order of probabilities and the most probable ones
    are selected from them, the number of which is set by the "beam_width" parameter.

    """

    step_candidates = list()

    for chain, score in candidates:
        prediction = model.predict(chain, encoded_src, tgt_attn_mask=None, tgt_pad_mask=None, src_pad_mask=None)
        probabilities = F.log_softmax(prediction[0, -1], dim=-1) + step_mask

        values, indices = probabilities.topk(k=min(beam_width, step_width))
        for prob, pos in zip(values, indices):
            new_token = torch.zeros(1, 1, model.token_size, device=device)
            new_token[0, 0, pos] = 1

            step_candidates.append((torch.cat([chain, new_token], dim=1), score + float(prob)))

    return sorted(step_candidates, key=lambda candidate: -candidate[1])[:beam_width]


def celery_task_loop(task: celery.Task
                     ) -> Callable[[Tensor, Tensor, TRecover, int, torch.device], List[Tuple[Tensor, float]]]:
    """
    Get a beam search algorithm loop function, which is implemented for the Celery task.

    Parameters
    ----------
    task : celery.Task
        Celery task base class.

    Returns
    -------
    inner_loop : Callable[[Tensor, Tensor, TRecover, int, torch.device], List[Tuple[Tensor, float]]]
        Beam search algorithm loop function for the Celery task.

    """

    def inner_loop(src: Tensor,
                   encoded_src: Tensor,
                   model: TRecover,
                   width: int,
                   device: torch.device
                   ) -> List[Tuple[Tensor, float]]:
        """
        Beam search algorithm loop implementation for the Celery task.

        Parameters
        ----------
        src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
            Keyless reading columns that are passed to the TRecover encoder.
        encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
            Keyless reading columns that were encoded by TRecover encoder.
        model : TRecover
            Trained model for keyless reading.
        width : int
            Number of candidates that can be selected at each step.
        device : torch.device
            Device on which to allocate the candidate chains.

        Returns
        -------
        candidates : List[Tuple[Tensor[1, SEQUENCE_LEN + 1, TOKEN_SIZE], float]]
            List of chains sorted in descending order of probabilities.
            The number of candidates is set by the "width" parameter.

        Notes
        -----
        Probability masks and beam widths values required for each step of the
        beam search algorithm are calculated by the "get_steps_params" function using
        keyless reading columns ("src") that are passed to the TRecover encoder.

        An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
        is used as the first candidate for the algorithm.

        The progress of the Celery task is updated at each step of the algorithm.

        """

        step_masks, step_widths = get_steps_params(src)
        candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

        for i in range(encoded_src.shape[0]):
            candidates = beam_step(candidates, step_masks[i], step_widths[i], encoded_src, model, width, device)
            task.update_state(meta={'progress': i + 1}, state='PREDICT')

        return candidates

    return inner_loop


def cli_interactive_loop(label: str = 'Processing'
                         ) -> Callable[[Tensor, Tensor, TRecover, int, torch.device], List[Tuple[Tensor, float]]]:
    """
    Get a beam search algorithm loop function, which is implemented for the cli interface.

    Parameters
    ----------
    label : str
        Label for the cli progress bar.

    Returns
    -------
    inner_loop : Callable[[Tensor, Tensor, TRecover, int, torch.device], List[Tuple[Tensor, float]]]
        Beam search algorithm loop function for the cli interface.

    """

    def inner_loop(src: Tensor,
                   encoded_src: Tensor,
                   model: TRecover,
                   width: int,
                   device: torch.device
                   ) -> List[Tuple[Tensor, float]]:
        """
        Beam search algorithm loop implementation for the cli interface.

        Parameters
        ----------
        src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
            Keyless reading columns that are passed to the TRecover encoder.
        encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
            Keyless reading columns that were encoded by TRecover encoder.
        model : TRecover
            Trained model for keyless reading.
        width : int
            Number of candidates that can be selected at each step.
        device : torch.device
            Device on which to allocate the candidate chains.

        Returns
        -------
        candidates : List[Tuple[Tensor[1, SEQUENCE_LEN + 1, TOKEN_SIZE], float]]
            List of chains sorted in descending order of probabilities.
            The number of candidates is set by the "width" parameter.

        Notes
        -----
        Probability masks and beam widths values required for each step of the
        beam search algorithm are calculated by the "get_steps_params" function using
        keyless reading columns ("src") that are passed to the TRecover encoder.

        An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
        is used as the first candidate for the algorithm.

        At each step of the algorithm, the progress bar of the task
        is updated and displayed in the console.

        """

        step_masks, step_widths = get_steps_params(src)
        candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

        with Progress(
                TextColumn('{task.description}', style='bright_blue'),
                BarColumn(complete_style='bright_blue'),
                TextColumn('{task.percentage:>3.0f}%', style='bright_blue'),
                TextColumn('Remaining', style='bright_blue'),
                TimeRemainingColumn(),
                TextColumn('Elapsed', style='bright_blue'),
                TimeElapsedColumn(),
                transient=True,
                console=log.project_console
        ) as progress:
            beam_progress = progress.add_task(label, total=encoded_src.shape[0])

            for i in range(encoded_src.shape[0]):
                candidates = beam_step(candidates, step_masks[i], step_widths[i], encoded_src, model, width, device)
                progress.update(beam_progress, advance=1)

        return candidates

    return inner_loop


def dashboard_loop(src: Tensor,
                   encoded_src: Tensor,
                   model: TRecover,
                   width: int,
                   device: torch.device
                   ) -> List[Tuple[Tensor, float]]:
    """
    Beam search algorithm loop implementation for the dashboard interface.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.
    encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
        Keyless reading columns that were encoded by TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    width : int
        Number of candidates that can be selected at each step.
    device : torch.device
        Device on which to allocate the candidate chains.

    Returns
    -------
    candidates : List[Tuple[Tensor[1, SEQUENCE_LEN + 1, TOKEN_SIZE], float]]
        List of chains sorted in descending order of probabilities.
        The number of candidates is set by the "width" parameter.

    Notes
    -----
    Probability masks and beam widths values required for each step of the
    beam search algorithm are calculated by the "get_steps_params" function using
    keyless reading columns ("src") that are passed to the TRecover encoder.

    An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
    is used as the first candidate for the algorithm.

    At each step of the algorithm, the progress bar is updated and displayed on the dashboard.

    """

    import streamlit as st

    step_masks, step_widths = get_steps_params(src)
    candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

    progress = st.progress(0)

    for i in range(encoded_src.shape[0]):
        candidates = beam_step(candidates, step_masks[i], step_widths[i], encoded_src, model, width, device)
        progress.progress(i / encoded_src.shape[0])

    return candidates


def standard_loop(src: Tensor,
                  encoded_src: Tensor,
                  model: TRecover,
                  width: int,
                  device: torch.device
                  ) -> List[Tuple[Tensor, float]]:
    """
    Base implementation of the beam search algorithm loop.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.
    encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
        Keyless reading columns that were encoded by TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    width : int
        Number of candidates that can be selected at each step.
    device : torch.device
        Device on which to allocate the candidate chains.

    Returns
    -------
    candidates : List[Tuple[Tensor[1, SEQUENCE_LEN + 1, TOKEN_SIZE], float]]
        List of chains sorted in descending order of probabilities.
        The number of candidates is set by the "width" parameter.

    Notes
    -----
    Probability masks and beam widths values required for each step of the
    beam search algorithm are calculated by the "get_steps_params" function using
    keyless reading columns ("src") that are passed to the TRecover encoder.

    An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
    is used as the first candidate for the algorithm.

    """

    step_masks, step_widths = get_steps_params(src)
    candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

    for i in range(encoded_src.shape[0]):
        candidates = beam_step(candidates, step_masks[i], step_widths[i], encoded_src, model, width, device)

    return candidates


def beam_search(src: Tensor,
                model: TRecover,
                width: int,
                device: torch.device,
                beam_loop: Callable[[Tensor, Tensor, TRecover, int, torch.device],
                                    List[Tuple[Tensor, float]]] = standard_loop
                ) -> List[Tuple[Tensor, float]]:
    """
    Beam search algorithm implementation.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    width : int
        Number of candidates that can be selected at each step.
    device : torch.device
        Device on which to allocate the candidate chains.
    beam_loop : Callable[[Tensor, Tensor, TRecover, int, torch.device],
                        List[Tuple[Tensor, float]]], default=standard_loop
        Beam search algorithm loop function.

    Returns
    -------
    candidates : List[Tuple[Tensor[SEQUENCE_LEN], float]]
        List of chains sorted in descending order of probabilities.
        The number of candidates is set by the "width" parameter.

    Notes
    -----
    Initially, the keyless reading columns ("src") are encoded using TRecover encoder,
    then the encoded columns ("encoded_src") are used at each step of the algorithm.

    """

    encoded_src = model.encode(src.unsqueeze(dim=0), src_pad_mask=None)

    candidates = beam_loop(src, encoded_src, model, width, device)

    return [
        (torch.argmax(chain.squeeze(), dim=-1)[1:], score)  # first token is empty_token
        for chain, score in candidates
    ]


# ---------------------------------------------Asynchronous Beam Search-------------------------------------------------


async def async_beam_step(candidates: List[Tuple[Tensor, float]],
                          step_mask: Tensor,
                          step_width: int,
                          encoded_src: Tensor,
                          model: TRecover,
                          beam_width: int,
                          device: torch.device
                          ) -> List[Tuple[Tensor, float]]:
    """
    Asynchronous implementation of the beam search algorithm step.

    Parameters
    ----------
    candidates : List[Tuple[Tensor[1, STEP_NUMBER, TOKEN_SIZE], float]]
        List of candidates from the previous step.
    step_mask : Tensor[TOKEN_SIZE]
        Column's mask that consists of zeros in the places that correspond to the letters allowed
        for selection in the column and values equal to minus infinity in all others.
        Required so that only the letters in the column are selected as a candidates.
    step_width : int
        Number of candidates that are contained in the step column.
    encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
        Columns for keyless reading that were encoded by TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    beam_width : int
        Number of candidates that can be selected at the current step.
    device : torch.device
        Device on which to allocate the candidate chains.

    Returns
    -------
    step_candidates : List[Tuple[Tensor[1, STEP_NUMBER + 1, TOKEN_SIZE], float]]
        List of candidates of size "beam_width" for the current step
        sorted in descending order of their probabilities.

    Notes
    -----
    For each chain candidate from the previous step:
    *       Probability distribution is calculated using trained TRecover model to
            select the next symbol from the current column,taking into account the "step_mask".
    *       The most probable symbols are selected from the calculated probability distribution,
            the number of which is set by the "step_width" and "beam_width" parameters.
    *       For each selected symbol, a new candidate chain with updated probability
            is constructed and placed in the "step_candidates" list.

    All candidates are sorted in descending order of probabilities and the most probable ones
    are selected from them, the number of which is set by the "beam_width" parameter.

    """

    async def candidate_step(chain: Tensor, score: float) -> None:
        prediction = model.predict(chain, encoded_src, tgt_attn_mask=None, tgt_pad_mask=None, src_pad_mask=None)
        probabilities = F.log_softmax(prediction[0, -1], dim=-1) + step_mask

        values, indices = probabilities.topk(k=min(beam_width, step_width))
        for prob, pos in zip(values, indices):
            new_token = torch.zeros(1, 1, model.token_size, device=device)
            new_token[0, 0, pos] = 1

            step_candidates.append((torch.cat([chain, new_token], dim=1), score + float(prob)))

    step_candidates = list()

    for candidate_chain, candidate_score in candidates:
        await candidate_step(candidate_chain, candidate_score)

    return sorted(step_candidates, key=lambda candidate: -candidate[1])[:beam_width]


def api_interactive_loop(queue: asyncio.Queue,
                         delimiter: str = ''
                         ) -> Callable[[Tensor, Tensor, TRecover, int, torch.device], Awaitable]:
    """
    Get an asynchronous beam search algorithm loop function, which is implemented for the API interface.

    Parameters
    ----------
    queue : asyncio.Queue
        Asynchronous queue for storing intermediate results.
    delimiter: str, default=''
        Delimiter for columns visualization.

    Returns
    -------
    async_inner_loop : Callable[[Tensor, Tensor, TRecover, int, torch.device], Awaitable]
        Asynchronous beam search algorithm loop function for the API interface.

    """

    async def async_inner_loop(src: Tensor,
                               encoded_src: Tensor,
                               model: TRecover,
                               width: int,
                               device: torch.device
                               ) -> None:
        """
        Asynchronous beam search algorithm loop implementation for the API interface.

        Parameters
        ----------
        src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
            Keyless reading columns that are passed to the TRecover encoder.
        encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
            Keyless reading columns that were encoded by TRecover encoder.
        model : TRecover
            Trained model for keyless reading.
        width : int
            Number of candidates that can be selected at each step.
        device : torch.device
            Device on which to allocate the candidate chains.

        Notes
        -----
        Probability masks and beam widths values required for each step of the
        beam search algorithm are calculated by the "get_steps_params" function using
        keyless reading columns ("src") that are passed to the TRecover encoder.

        An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
        is used as the first candidate for the algorithm.

        At each step of the algorithm, the intermediate results are placed in an asynchronous queue.

        At the end of the algorithm, a None value is placed in the asynchronous queue,
        which is an indicator of its completion.

        """

        step_masks, step_widths = get_steps_params(src)
        candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

        for i in range(encoded_src.shape[0]):
            candidates = await async_beam_step(candidates, step_masks[i], step_widths[i],
                                               encoded_src, model, width, device)

            intermediate_result = [
                (tensor_to_target(torch.argmax(chain.squeeze(), dim=-1)[1:]), score)
                for chain, score in candidates
            ]

            await queue.put([(visualize_target(chain, delimiter), score) for chain, score in intermediate_result])

        await queue.put(None)

    return async_inner_loop


async def standard_async_loop(src: Tensor,
                              encoded_src: Tensor,
                              model: TRecover,
                              width: int,
                              device: torch.device
                              ) -> List[Tuple[Tensor, float]]:
    """
    Base asynchronous implementation of the beam search algorithm loop.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.
    encoded_src : Tensor[SEQUENCE_LEN, 1, D_MODEL]
        Keyless reading columns that were encoded by TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    width : int
        Number of candidates that can be selected at each step.
    device : torch.device
        Device on which to allocate the candidate chains.

    Returns
    -------
    candidates : List[Tuple[Tensor[1, SEQUENCE_LEN + 1, TOKEN_SIZE], float]]
        List of chains sorted in descending order of probabilities.
        The number of candidates is set by the "width" parameter.

    Notes
    -----
    Probability masks and beam widths values required for each step of the
    beam search algorithm are calculated by the "get_steps_params" function using
    keyless reading columns ("src") that are passed to the TRecover encoder.

    An empty chain (zero token of shape [1, 1, TOKEN_SIZE]) with zero probability
    is used as the first candidate for the algorithm.

    """

    step_masks, step_widths = get_steps_params(src)
    candidates = [(torch.zeros(1, 1, model.token_size, device=device), 0)]

    for i in range(encoded_src.shape[0]):
        candidates = await async_beam_step(candidates, step_masks[i], step_widths[i],
                                           encoded_src, model, width, device)

    return candidates


async def async_beam_search(src: Tensor,
                            model: TRecover,
                            width: int,
                            device: torch.device,
                            beam_loop: Callable[[Tensor, Tensor, TRecover, int, torch.device],
                                                Awaitable[Optional[List[Tuple[Tensor, float]]]]] = standard_async_loop
                            ) -> Optional[List[Tuple[Tensor, float]]]:
    """
    Asynchronous beam search algorithm implementation.

    Parameters
    ----------
    src : Tensor[SEQUENCE_LEN, TOKEN_SIZE]
        Keyless reading columns that are passed to the TRecover encoder.
    model : TRecover
        Trained model for keyless reading.
    width : int
        Number of candidates that can be selected at each step.
    device : torch.device
        Device on which to allocate the candidate chains.
    beam_loop : Callable[[Tensor, Tensor, TRecover, int, torch.device],
                        List[Tuple[Tensor, float]]], default=standard_loop
        Beam search algorithm loop function.

    Returns
    -------
    candidates : Optional[List[Tuple[Tensor[SEQUENCE_LEN], float]]]
        List of chains sorted in descending order of probabilities.
        The number of candidates is set by the "width" parameter.
        Returns None if "api_interactive_loop" is used as a beam search loop function.

    Notes
    -----
    Initially, the keyless reading columns ("src") are encoded using TRecover encoder,
    then the encoded columns ("encoded_src") are used at each step of the asynchronous algorithm.

    """

    encoded_src = model.encode(src.unsqueeze(dim=0), src_pad_mask=None)

    candidates = await beam_loop(src, encoded_src, model, width, device)

    return [(torch.argmax(chain.squeeze(), dim=-1)[1:], score) for chain, score in candidates] if candidates else None



import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import CrossEntropyLoss


class CustomCrossEntropyLoss(torch.nn.Module):
    # TODO docs

    def __init__(self, ignore_index: int = -1):
        super(CustomCrossEntropyLoss, self).__init__()

        self.criterion = CrossEntropyLoss(ignore_index=ignore_index)

    def forward(self, *tensors) -> Tensor:
        tgt, tgt_out = tensors[2], tensors[6]

        return self.criterion(tgt_out, tgt)


class CustomPenaltyLoss(torch.nn.Module):
    # TODO docs

    def __init__(self, coefficient: float = 1.0, ignore_index: int = -1):
        super(CustomPenaltyLoss, self).__init__()

        self.coefficient = coefficient
        self.criterion = CrossEntropyLoss(ignore_index=ignore_index)

    def forward(self, *tensors) -> Tensor:
        src, _, tgt, src_pad_mask, _, _, tgt_out = tensors

        penalty_tensor = -1.0 * F.log_softmax(tgt_out, dim=-1) * ~src.bool()

        if isinstance(src_pad_mask, Tensor):
            penalty_tensor = torch.masked_select(penalty_tensor, mask=~src_pad_mask.view(-1, 1))

        return self.criterion(tgt_out, tgt) + self.coefficient * penalty_tensor.mean()



from typing import Optional, Dict

import celery

from trecover.config import var


class ModelConfigTask(celery.Task):
    def __init__(self):
        super(ModelConfigTask, self).__init__()

        self.config: Optional[Dict] = None

    def __call__(self, *args, **kwargs):
        """
        Load params config on first call (i.e. first task processed).
        Avoids the need to load params config on each task request.

        """

        if not self.config:
            import torch
            from trecover.utils.model import load_params

            config = load_params(var.INFERENCE_PARAMS_PATH)
            config.cuda = var.CUDA and torch.cuda.is_available()

            self.config = config.jsonify()

        return self.run(*args, **kwargs)


class PredictTask(celery.Task):
    def __init__(self):

        super(PredictTask, self).__init__()

        self.device = None
        self.model = None

    def __call__(self, *args, **kwargs):
        """
        Load model on first call (i.e. first task processed).
        Avoids the need to load model on each task request.

        """

        if not self.device:
            import torch

            self.device = torch.device(f'cuda' if var.CUDA and torch.cuda.is_available() else 'cpu')

        if not self.model:
            from trecover.utils.model import get_model, load_params

            self.update_state(state='LOADING')

            params = load_params(var.INFERENCE_PARAMS_PATH)
            self.model = get_model(params.token_size, params.pe_max_len, params.num_layers,
                                   params.d_model, params.n_heads, params.d_ff,
                                   params.dropout, self.device, weights=var.INFERENCE_WEIGHTS_PATH, silently=True)
            self.model.eval()

        self.update_state(state='PREDICT')

        return self.run(*args, **kwargs)



from pathlib import Path

import typer
from typer import Typer, Argument, Option, Context

from trecover.app.cli import download, train, mlflow, dashboard, api, worker, broker, backend
from trecover.config import var

cli = Typer(name='TRecover-cli', add_completion=False)

cli.add_typer(download.cli, name='download')
cli.add_typer(train.cli, name='train')
cli.add_typer(mlflow.cli, name='mlflow')
cli.add_typer(dashboard.cli, name='dashboard')
cli.add_typer(api.cli, name='api')
cli.add_typer(worker.cli, name='worker')
cli.add_typer(broker.cli, name='broker')
cli.add_typer(backend.cli, name='backend')


@cli.command(help='Perform keyless reading')
def recover(data_path: Path = Argument(..., help='Path to file or dir for data', exists=True),
            model_params: Path = Option(var.INFERENCE_PARAMS_PATH, help='Path to model params json file', exists=True),
            weights_path: Path = Option(var.INFERENCE_WEIGHTS_PATH, help='Path to model weights', exists=True),
            cuda: bool = Option(var.CUDA, envvar='CUDA', help='CUDA enabled'),
            gpu_id: int = Option(0, help='GPU id'),
            separator: str = Option(' ', help='Columns separator in the input files'),
            noisy: bool = Option(False, help='Input files are noisy texts'),
            min_noise: int = Option(3, help='Min noise parameter. Minimum value is zero'),
            max_noise: int = Option(5, help='Max noise parameter. Maximum value is alphabet size'),
            beam_width: int = Option(5, help='Width for beam search algorithm. Maximum value is alphabet size'),
            n_to_show: int = Option(0, help="Number of columns to visualize. Zero value means for no restriction's"),
            delimiter: str = Option('', help='Delimiter for columns visualization')
            ) -> None:
    """
    Perform keyless reading locally.

    Parameters
    ----------
    data_path : Path
        Path to file or dir for data.
    model_params : Path
        Path to model params json file.
    weights_path : Path
        Path to model weights.
    cuda : bool
        CUDA enabled.
    gpu_id : int, default=0
        GPU id on which perform computations.
    separator : str, default=' '
        Columns separator in the input files.
    noisy : bool, default=False
        Indicates that input files are noisy texts.
    min_noise : int, default=3
        Min noise size per column. Minimum value is zero.
    max_noise : int, default=5
        Max noise size per column. Maximum value is alphabet size.
    beam_width : int, default=5
        Width for beam search algorithm. Maximum value is alphabet size.
    n_to_show : int, default=0
        Number of columns to visualize. Zero value means for no restriction's.
    delimiter : str, default=''
        Delimiter for columns visualization.

    Examples
    --------
    >>> trecover recover examples/example_1.txt
    ╭──────────────────────────────────────────────────── example_1.txt ───────────────────────────────────────────────╮
    │                                                        Columns                                                   │
    │ ajocmbfeafodadbddciafqnahdfeihhkieeaacacafkdchddakhecmmlibfinaehbcbdiicejkeahnfemaeaadbkagacbdmahbibacfddfbbbca… │
    │ enpenkhgglrifflheioentrmjenkjnrmlhphdddeihliekeeeolflonpmctjolgkdeljjmljmmjiisjknjghgeelhkbddlpjjekrkdkilgiocii… │
    │ gsxtoplqkrtknksinktipwvnlnqqrstotoqspoejtsnoiuoflpohvtovqeutunjojlmksonosskpvxporrltnfgoprdemstnshnssgnronjreqj… │
    │ xvzwttqtxvxuoptowuxnxyzrwrrtwtyqwqvutrwrxvtxxwurrtqlwuqzvnwvxossmmpnutosuxlswyuvtttvqulrqzrrwuxtyqouwiuupwsxnrm… │
    │  y y yz zy  y w zy uz  yys   u tzs   x u         wx     wy w tuvpuwu  x yyowyz  z  wxyu     xyy   v yr    t yvw… │
    │                                                       Predicted                                                  │
    │ enpeoplearoundthecountrywereintothestreetstickedatheconvictionsspewditnessesinpentlandboardeddytheirwindowsbyra… │
    │                                                                                                                  │
    ╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯
    Elapsed:   4.716 s

    Notes
    -----
    A larger "beam_width" parameter value can improve keyless reading, but it will also take longer to compute.

    """

    from time import time

    import torch
    from rich.console import Group
    from rich.panel import Panel
    from rich.progress import Progress, TextColumn
    from rich.text import Text

    from trecover.config import log
    from trecover.utils.beam_search import beam_search, cli_interactive_loop
    from trecover.utils.cli import get_files_columns
    from trecover.utils.transform import files_columns_to_tensors
    from trecover.utils.model import get_model, load_params
    from trecover.utils.transform import tensor_to_columns, tensor_to_target
    from trecover.utils.visualization import visualize_columns, visualize_target

    data_path = Path(data_path)
    params = load_params(Path(model_params))

    if not noisy and min_noise >= max_noise:
        log.project_logger.error('[red]Maximum noise range must be grater than minimum noise range')
        return

    if not any([data_path.is_file(), data_path.is_dir()]):
        log.project_logger.error('[red]Files for inference needed to be specified')
        return

    if params.pe_max_len < n_to_show:
        log.project_logger.error(f'[red]Parameter n_to_show={n_to_show} must be less than {params.pe_max_len}')
        return
    elif n_to_show == 0:
        n_to_show = params.pe_max_len

    device = torch.device(f'cuda:{gpu_id}' if cuda and torch.cuda.is_available() else 'cpu')

    with Progress(TextColumn('{task.description}', style='bright_blue'),
                  transient=True,
                  console=log.project_console
                  ) as progress:
        progress.add_task('Model loading...')
        model = get_model(params.token_size, params.pe_max_len, params.num_layers,
                          params.d_model, params.n_heads, params.d_ff, params.dropout,
                          device, weights=Path(weights_path), silently=True)
    model.eval()

    files, files_columns = get_files_columns(data_path, separator, noisy, min_noise, max_noise, n_to_show)
    files_src = files_columns_to_tensors(files_columns, device)

    for file_id, (file, src) in enumerate(zip(files, files_src), start=1):
        start_time = time()

        loop_label = f'{file_id}/{len(files_src)} Processing {file.name}'
        chains = beam_search(src, model, beam_width, device,
                             beam_loop=cli_interactive_loop(label=loop_label))
        chains = [Text(visualize_target(tensor_to_target(chain), delimiter=delimiter), style='cyan', justify='center',
                       overflow='ellipsis', end='\n\n') for (chain, _) in chains]

        columns = visualize_columns(tensor_to_columns(src), delimiter=delimiter, as_rows=True)
        columns = (Text(row, style='bright_blue', overflow='ellipsis', no_wrap=True, justify='left') for row in columns)

        panel_group = Group(
            Text('Columns', style='magenta', justify='center'),
            *columns,
            Text('Predicted', style='magenta', justify='center'),
            *chains
        )

        log.project_console.print(
            Panel(panel_group, title=file.name, border_style='magenta'),
            justify='center'
        )

        log.project_console.print(f'\nElapsed: {time() - start_time:>7.3f} s\n', style='bright_blue')


@cli.callback(invoke_without_command=True, help='')
def cli_state_verification(ctx: Context,
                           config_file: Path = Option(var.BASE_DIR / 'trecover-compose.toml', '--file', '-f',
                                                      file_okay=True,
                                                      help='Path to TRecover configuration file for "up" command'),
                           attach_stream: bool = Option(False, '--attach', '-a', is_flag=True,
                                                        help='Attach output and error streams for "up" command')
                           ) -> None:
    """
    Perform cli commands verification (state checking) and config file parsing.

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.
    config_file : Path, default=var.BASE_DIR / 'trecover-compose.toml'
        Path to TRecover configuration file for "up" command.
    attach_stream : bool, default=False
        Attach output and error streams for "up" command'.

    """

    if ctx.invoked_subcommand == 'init':
        return
    elif not var.BASE_INIT.exists():
        typer.echo(typer.style('You need to initialize the project environment.\n'
                               'For more information use: trecover init --help',
                               fg=typer.colors.RED))
        ctx.exit(1)

    from trecover.config import log

    if ctx.invoked_subcommand is None:
        log.project_console.print(ctx.get_help(), markup=False)
        ctx.exit(0)

    if ctx.invoked_subcommand in ('up', 'down', 'status', 'attach'):
        from trecover.utils.docker import is_docker_running
        from trecover.utils.cli import parse_config

        if not is_docker_running():
            log.project_console.print('Docker engine is not running', style='red')
            ctx.exit(1)

        if not config_file.exists():
            log.project_console.print('Defined TRecover configuration file does not exist', style='red')
            ctx.exit(1)

        ctx.params['conf'] = parse_config(config_file)
        ctx.params['attach'] = attach_stream


@cli.command(help="Initialize project's environment")
def init(base: Path = Option(Path().absolute(), '--base', '-b', help="Path to the project's base directory"),
         relocate: bool = Option(False, '--relocate', '-r', is_flag=True, help='Relocate an existing environment')
         ) -> None:
    """
    Initialize project's environment.

    Parameters
    ----------
    base : Path, default='./'
        Path to the project's base directory.
    relocate : bool, default=False
        Relocate an existing environment

    """

    from shutil import move, Error

    def rebase(src: Path, dst: Path) -> None:
        try:
            move(str(src), str(dst))
        except (PermissionError, Error):
            typer.echo(typer.style(f'Failed to relocate: {src}', fg=typer.colors.YELLOW))
        else:
            typer.echo(typer.style(f'Relocated: {src}', fg=typer.colors.BRIGHT_BLUE))

    base.mkdir(parents=True, exist_ok=True)

    with var.BASE_INIT.open(mode='w') as f:
        f.write(base.as_posix())

    typer.echo(typer.style("Project's environment is initialized.", fg=typer.colors.BRIGHT_BLUE))

    if relocate:
        if var.LOGS_DIR.exists():
            rebase(var.LOGS_DIR, base)
        if var.INFERENCE_DIR.exists():
            rebase(var.INFERENCE_DIR, base)
        if var.DATA_DIR.exists():
            rebase(var.DATA_DIR, base)
        if var.EXPERIMENTS_DIR.exists():
            rebase(var.EXPERIMENTS_DIR, base)


@cli.command(help='Start services')
def up(ctx: Context) -> None:
    """
    Start services: Dashboard, API, Worker, Broker, Backend.

    Command uses trecover-compose.toml config file specified by --file and
    attaches output and error streams if --attach flag is set.

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

    Config Variables
    ----------------
    dashboard host : str, default=ENV(STREAMLIT_HOST) or 'localhost'
        The address where the server will listen for client and browser connections.
        Use this if you want to bind the server to a specific address. If set, the server
        will only be accessible from this address, and not from any aliases (like localhost).

    dashboard port : int, default=ENV(STREAMLIT_PORT) or 8000
        The port where the server will listen for browser connections.

    dashboard loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.

    api host : str, default=ENV(FASTAPI_HOST) or 'localhost'
        Bind socket to this host.

    api port : int, default=ENV(FASTAPI_PORT) or 8001
        Bind socket to this port.

    api loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.

    api concurrency : int, default=ENV(FASTAPI_WORKERS) or 1
        The number of worker processes.

    worker name : str, default='TRecoverWorker'
        Custom worker hostname.

    worker pool : str, {'prefork', 'eventlet', 'gevent', 'processes', 'solo'}, default='solo'
        Worker processes/threads pool type.

    worker loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.

    worker concurrency : int, default=ENV(CELERY_WORKERS) or 1
        The number of worker processes.

    worker broker_url : str, default=ENV(CELERY_BROKER) or 'pyamqp://guest@localhost'
        Broker url.

    worker backend_url : str, default=ENV(CELERY_BACKEND) or 'redis://localhost'
        Backend url.

    broker port : int, default=ENV(BROKER_PORT) or 5672
        Bind broker socket to this port.

    broker ui_port : int, default=ENV(BROKER_UI_PORT) or 15672
        Bind UI socket to this port.

    broker auto_remove : bool, default=False
        Remove broker docker container after service exit.

    backend port : int, default=ENV(BACKEND_PORT) or 6379
        Bind backend socket to this port.

    backend auto_remove : bool, default=False
        Remove backend docker container after service exit.

    Examples
    --------
    # trecover-compose.toml

        [dashboard]
        host = "localhost"
        port = 8000
        loglevel = 'info'

        [api]
        host = "localhost"
        port = 8001
        loglevel = 'info'
        concurrency = 1

        [worker]
        name = "TRecoverWorker"
        pool = "solo"
        loglevel = "info"
        concurrency = 1
        broker_url = "pyamqp://guest@localhost:5672"
        backend_url = "redis://localhost:6379"

        [broker]
        port = 5672
        ui_port = 15672
        auto_remove = false

        [backend]
        port = 6379
        auto_remove = false

    """

    from trecover.config import log
    from trecover.utils.docker import get_container
    from trecover.utils.cli import check_service

    conf = ctx.parent.params['conf']

    if var.DASHBOARD_PID.exists():
        check_service(name='dashboard', pidfile=var.DASHBOARD_PID)
    else:
        dashboard.dashboard_start(host=conf.dashboard.host,
                                  port=conf.dashboard.port,
                                  loglevel=conf.dashboard.loglevel,
                                  attach=False,
                                  no_daemon=False)

    if var.API_PID.exists():
        check_service(name='API', pidfile=var.API_PID)
    else:
        api.api_start(host=conf.api.host,
                      port=conf.api.port,
                      loglevel=conf.api.loglevel,
                      concurrency=conf.api.concurrency,
                      attach=False,
                      no_daemon=False)

    if var.WORKER_PID.exists():
        check_service(name='worker', pidfile=var.WORKER_PID)
    else:
        worker.worker_start(name=conf.worker.name,
                            pool=conf.worker.pool,
                            loglevel=conf.worker.loglevel,
                            concurrency=conf.worker.concurrency,
                            broker_url=conf.worker.broker_url,
                            backend_url=conf.worker.backend_url,
                            attach=False,
                            no_daemon=False)

    if (container := get_container(var.BROKER_ID)) and container.status == 'running':
        log.project_console.print(':rocket: The broker status: running', style='bright_blue')
    else:
        broker.broker_start(port=conf.broker.port,
                            ui_port=conf.broker.ui_port,
                            auto_remove=conf.broker.auto_remove,
                            attach=False)

    if (container := get_container(var.BACKEND_ID)) and container.status == 'running':
        log.project_console.print(':rocket: The backend status: running', style='bright_blue')
    else:
        backend.backend_start(port=conf.backend.port,
                              auto_remove=conf.backend.auto_remove,
                              attach=False)

    if ctx.parent.params['attach']:
        try:
            attach(live=False)
        finally:
            down(prune=False, v=False)


@cli.command(help='Stop services')
def down(prune: bool = Option(False, '--prune', '-p', is_flag=True,
                              help='Prune all docker containers after exit.'),
         v: bool = Option(False, '--volume', '-v', is_flag=True,
                          help='Remove the volumes associated with the all docker containers and the log files.')
         ) -> None:
    """
    Stop services: Dashboard, API, Worker, Broker, Backend.

    Parameters
    ----------
    prune : bool, default=False
        Prune all docker containers after exit.
    v : bool, default=False
        Remove the volumes associated with the all docker containers and the log files.

    """

    from trecover.config import log
    from trecover.utils.docker import get_container
    from trecover.utils.cli import stop_service

    if var.DASHBOARD_PID.exists():
        stop_service(name='dashboard', pidfile=var.DASHBOARD_PID, logfile=log.DASHBOARD_LOG)

    if var.API_PID.exists():
        stop_service(name='API', pidfile=var.API_PID, logfile=log.API_LOG)

    if var.WORKER_PID.exists():
        stop_service(name='worker', pidfile=var.WORKER_PID, logfile=log.WORKER_LOG)

    if (container := get_container(var.BROKER_ID)) and container.status == 'running':
        container.stop()

        log.project_console.print('The broker service is stopped', style='bright_blue')

        if prune:
            broker.broker_prune(force=False, v=v)

    if (container := get_container(var.BACKEND_ID)) and container.status == 'running':
        container.stop()

        log.project_console.print('The backend service is stopped', style='bright_blue')

        if prune:
            backend.backend_prune(force=False, v=v)


@cli.command(name='status', help='Display services status')
def status() -> None:
    """ Display services status """

    from trecover.utils.cli import check_service
    from trecover.app.cli.broker import broker_status
    from trecover.app.cli.backend import backend_status

    check_service(name='dashboard', pidfile=var.DASHBOARD_PID)
    check_service(name='API', pidfile=var.API_PID)
    check_service(name='worker', pidfile=var.WORKER_PID)
    broker_status()
    backend_status()


@cli.command(name='attach', help='Attach local output stream to a running services')
def attach(live: bool = Option(False, '--live', '-l', is_flag=True,
                               help='Stream only fresh log records')
           ) -> None:
    """
    Attach local output stream to a running services.

    Parameters
    ----------
    live : bool, Default=False
        Stream only fresh log records

    """

    from trecover.config import log
    from trecover.utils.cli import stream

    with log.project_console.screen():
        for record in stream(('dashboard', log.DASHBOARD_LOG),
                             ('API', log.API_LOG),
                             ('worker', log.WORKER_LOG),
                             live=live):
            log.project_console.print(record)

    log.project_console.clear()


if __name__ == '__main__':
    cli()



import re
from pathlib import Path


def clean_wiki_text(filename: Path, drop_threshold: float = 0.62) -> None:
    result_size = 0

    with open(filename, mode='r') as wiki_file, open(f'{filename}.clean', mode='w') as writer:

        while True:
            try:
                line = wiki_file.readline()
            except UnicodeDecodeError:
                continue

            if not line:
                break

            if not line.strip() or line.startswith(' = '):
                continue

            cleaned_line = re.sub(r'[^A-Za-z]', '', line)

            if len(cleaned_line) / len(line) < drop_threshold:
                continue

            writer.write(cleaned_line.lower())
            result_size += len(cleaned_line)

    print(f'Result size: {result_size}')


def clean_wiki_qa(filename: Path) -> None:
    result_size = 0
    questions = set()

    with open(filename, mode='r') as wiki_file, open(f'{filename}.clean', mode='w') as writer:

        while True:
            try:
                line = wiki_file.readline()
            except UnicodeDecodeError:
                continue

            if not line:
                break

            try:
                question, answer = line.split('\t')[:2]
            except ValueError:
                continue

            cleaned_question = re.sub(r'[^A-Za-z]', '', question).lower()
            cleaned_answer = re.sub(r'[^A-Za-z]', '', answer).lower()

            if cleaned_question not in questions:
                writer.write(cleaned_question)

                questions.add(cleaned_question)
                result_size += len(cleaned_question)

            writer.write(cleaned_answer)
            result_size += len(cleaned_answer)

    print(f'Result size: {result_size}')


def clean_blogs(blogs_folder: Path) -> None:
    result_size = 0

    with open(blogs_folder / 'blogs.clean', mode='w') as writer:
        for blog_file in [Path(blogs_folder, file) for file in blogs_folder.iterdir()]:

            text_flag = False

            with open(blog_file, mode='r') as f:
                while True:
                    try:
                        line = f.readline()
                    except UnicodeDecodeError:
                        continue

                    if not line:
                        break

                    if text_flag and '</post>' in line:
                        text_flag = False

                    elif text_flag:
                        cleaned_text = re.sub(r'[^A-Za-z]', '', line).lower()

                        writer.write(cleaned_text)
                        result_size += len(cleaned_text)

                    elif not text_flag and '<post>' in line:
                        text_flag = True

    print(f'Result size: {result_size}')



# TODO docs

from argparse import ArgumentParser
from pathlib import Path
from typing import List, Optional

import torch

from trecover.config import var, exp_var, log
from trecover.train.data import WikiDataset, StandardCollate
from trecover.train.loss import CustomPenaltyLoss
from trecover.train.monitor import WandbMonitor, MlflowMonitor
from trecover.train.scheduler import WarmupScheduler
from trecover.train.trainer import LocalTrainer
from trecover.utils.model import get_model, get_recent_weights_path
from trecover.utils.train import ExperimentParams, set_seeds, get_experiment_mark


def get_parser() -> ArgumentParser:
    # TODO docs

    parser = ArgumentParser()

    # ------------------------------------------------GENERAL PARAMETERS------------------------------------------------

    parser.add_argument('--project-name', default='TRecoverLocal', type=str,
                        help='Monitor project name')
    parser.add_argument('--mlflow', action='store_true',
                        help='Use Mlflow as monitor. Default is W&B')

    # --------------------------------------------------DATA PARAMETERS-------------------------------------------------

    parser.add_argument('--seed', default=2531, type=int,
                        help='Reproducible seed number')
    parser.add_argument('--train-files', default=exp_var.TRAIN_DATA, type=str,
                        help='Path to train files folder')
    parser.add_argument('--val-files', default=exp_var.VAL_DATA, type=str,
                        help='Path to validation files folder')
    parser.add_argument('--vis-files', default=exp_var.VIS_DATA, type=str,
                        help='Path to visualization files folder')
    parser.add_argument('--test-files', default=exp_var.VIS_DATA, type=str,
                        help='Path to test files folder')
    parser.add_argument('--min-threshold', default=256, type=int,
                        help='Min sentence lengths')
    parser.add_argument('--max-threshold', default=256, type=int,
                        help='Max sentence lengths')
    parser.add_argument('--train-dataset-size', default=2000, type=int,
                        help='Train dataset size')
    parser.add_argument('--val-dataset-size', default=400, type=int,
                        help='Validation dataset size')
    parser.add_argument('--vis-dataset-size', default=5, type=int,
                        help='Visualization dataset size')
    parser.add_argument('--test-dataset-size', default=200, type=int,
                        help='Test dataset size')
    parser.add_argument('--batch-size', default=2, type=int,
                        help='Batch size')
    parser.add_argument('--n-workers', default=3, type=int,
                        help='Number of processes for dataloaders')
    parser.add_argument('--min-noise', default=0, type=int,
                        help='Min noise range')
    parser.add_argument('--max-noise', default=0, type=int,
                        help='Max noise range')
    parser.add_argument('--allocate-on-device', action='store_true',
                        help='Allocate train data on specified device during batch generation')

    # ------------------------------------------MODEL PARAMETERS--------------------------------------------------------

    parser.add_argument('--token-size', default=len(var.ALPHABET), type=int,
                        help='Token size')
    parser.add_argument('--pe-max-len', default=256, type=int,
                        help='Positional encoding max length')
    parser.add_argument('--n-layers', default=12, type=int,
                        help='Number of encoder and decoder blocks')
    parser.add_argument('--d-model', default=768, type=int,
                        help='Model dimension - number of expected features in the encoder (decoder) input')
    parser.add_argument('--n-heads', default=12, type=int,
                        help='Number of encoder and decoder attention heads')
    parser.add_argument('--d-ff', default=768, type=int,
                        help='Dimension of the feedforward layer')
    parser.add_argument('--dropout', default=0.1, type=float,
                        help='Dropout range')
    parser.add_argument('--exp-dir', default=var.EXPERIMENTS_DIR, type=str,
                        help='Experiments folder')
    parser.add_argument('--abs-weights-name', type=str,
                        help='Absolute weights path')
    parser.add_argument('--exp-mark', type=str, default='base',
                        help="Experiments folder mark placed in 'exp-dir'")
    parser.add_argument('--weights-name', type=str,
                        help="Weights name in specified using 'exp-mark' experiments folder")
    parser.add_argument('--no-cuda', action='store_true',
                        help='Disable cuda usage')

    # ----------------------------------------OPTIMIZATION PARAMETERS---------------------------------------------------

    parser.add_argument('--lr', default=0.001577, type=float,
                        help='Learning rate value.')
    parser.add_argument('--lr-step-seek', default=0, type=int,
                        help='Number of steps for WarmupScheduler to seek')
    parser.add_argument('--warmup', default=600, type=int,
                        help='Warmup value for WarmupScheduler')
    parser.add_argument('--lr-step-size', default=1, type=int,
                        help='Step size foe learning rate updating')
    parser.add_argument('--accumulation-step', default=1, type=int,
                        help='Number of steps for gradients accumulation')
    parser.add_argument('--penalty-coefficient', default=1.0, type=float,
                        help='Penalty coefficient for CustomPenaltyLoss')

    # -----------------------------------------TRAIN LOOP PARAMETERS----------------------------------------------------

    parser.add_argument('--n-epochs', default=1000, type=int,
                        help='Number of epochs for training')
    parser.add_argument('--epoch-seek', default=0, type=int,
                        help='Number of epochs to seek. necessary for correct weights naming'
                             ' in case of an interrupted model training process')
    parser.add_argument('--saving-interval', default=1, type=int,
                        help='Weights saving interval per epoch')
    parser.add_argument('--log-interval', default=1, type=int,
                        help='Metrics logging interval per batch-step')
    parser.add_argument('--vis-interval', default=1, type=int,
                        help='Visualization interval per epoch')
    parser.add_argument('--n-columns-to-show', default=96, type=int,
                        help='Number of visualization columns to show')
    parser.add_argument('--delimiter', default='', type=str,
                        help='Visualization columns delimiter')

    return parser


def get_experiment_params(args: Optional[List[str]] = None) -> ExperimentParams:
    # TODO docs

    return ExperimentParams(get_parser().parse_args(args=args))


def train(args: Optional[List[str]] = None) -> None:
    # TODO docs

    params = get_experiment_params(args)

    if params.n_columns_to_show > params.pe_max_len:
        log.project_logger.error(f'[red]Parameter n_to_show={params.n_columns_to_show} '
                                 f'must be less than {params.pe_max_len}')
        return

    set_seeds(seed=params.seed)

    device = torch.device('cuda' if torch.cuda.is_available() and not params.no_cuda else 'cpu')
    batch_generation_device = device if params.allocate_on_device else None
    weights_path = params.abs_weights_name or get_recent_weights_path(Path(params.exp_dir), params.exp_mark,
                                                                      params.weights_name)

    model = get_model(params.token_size, params.pe_max_len, params.n_layers, params.d_model, params.n_heads,
                      params.d_ff, params.dropout, device,
                      weights=weights_path, silently=False)

    train_files = [Path(params.train_files, file) for file in Path(params.train_files).iterdir()]
    val_files = [Path(params.val_files, file) for file in Path(params.val_files).iterdir()]
    vis_files = [Path(params.vis_files, file) for file in Path(params.vis_files).iterdir()]
    test_files = [Path(params.test_files, file) for file in Path(params.test_files).iterdir()]

    train_dataset = WikiDataset(datafiles=train_files, min_threshold=params.min_threshold,
                                max_threshold=params.max_threshold, dataset_size=params.train_dataset_size)
    val_dataset = WikiDataset(datafiles=val_files, min_threshold=params.min_threshold,
                              max_threshold=params.max_threshold, dataset_size=params.val_dataset_size)
    vis_dataset = WikiDataset(datafiles=vis_files, min_threshold=params.min_threshold,
                              max_threshold=params.max_threshold, dataset_size=params.vis_dataset_size)
    test_dataset = WikiDataset(datafiles=test_files, min_threshold=params.min_threshold,
                               max_threshold=params.max_threshold, dataset_size=params.test_dataset_size)

    collate = StandardCollate(min_noise=params.min_noise, max_noise=params.max_noise, device=batch_generation_device)

    train_loader = train_dataset.create_dataloader(batch_size=params.batch_size, collate=collate,
                                                   num_workers=params.n_workers)
    val_loader = val_dataset.create_dataloader(batch_size=params.batch_size, collate=collate,
                                               num_workers=params.n_workers)
    vis_loader = vis_dataset.create_dataloader(batch_size=params.batch_size, collate=collate,
                                               num_workers=params.n_workers)
    test_loader = test_dataset.create_dataloader(batch_size=params.batch_size, collate=collate,
                                                 num_workers=params.n_workers)

    # criterion = CustomCrossEntropyLoss(ignore_index=-1)
    criterion = CustomPenaltyLoss(coefficient=params.penalty_coefficient, ignore_index=-1)
    optimizer = torch.optim.Adam(model.parameters(), lr=params.lr, betas=(0.9, 0.98), eps=1e-9)

    scheduler = WarmupScheduler(optimizer, params.d_model, params.warmup, params.lr_step_size, seek=params.lr_step_seek)

    experiment_mark = get_experiment_mark()

    if params.mlflow:
        monitor = MlflowMonitor(params.project_name, experiment_mark, params.jsonify(),
                                exp_var.MLFLOW_REGISTRY_DIR.absolute().as_uri())
    else:
        monitor = WandbMonitor(params.project_name, experiment_mark, params.jsonify(),
                               exp_var.WANDB_REGISTRY_DIR.absolute())

    with LocalTrainer(params=params,
                      model=model,
                      criterion=criterion,
                      optimizer=optimizer,
                      exp_dir=Path(params.exp_dir),
                      scheduler=scheduler,
                      monitor=monitor,
                      device=device,
                      accumulation_step=params.accumulation_step,
                      log_interval=params.log_interval,
                      saving_interval=params.saving_interval,
                      vis_interval=params.vis_interval,
                      n_columns_to_show=params.n_columns_to_show,
                      delimiter=params.delimiter
                      ) as (trainer, _):
        trainer.train(params.n_epochs, train_loader, val_loader, vis_loader, params.epoch_seek)
        trainer.test(test_loader=test_loader)



from typing import Optional

from typer import Typer, Option, Context, BadParameter

from trecover.config import var

cli = Typer(name='Mlflow-cli', add_completion=False, help='Manage Mlflow service')


@cli.callback(invoke_without_command=True)
def mlflow_state_verification(ctx: Context) -> None:
    """
    Perform cli commands verification (state checking).

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

   """

    from trecover.config import log, exp_var

    if var.MLFLOW_PID.exists():
        if ctx.invoked_subcommand in ('start', None):
            log.project_console.print(':rocket: The mlflow service is already started', style='bright_blue')
            ctx.exit(0)

    elif ctx.invoked_subcommand is None:
        mlflow_start(host=var.MLFLOW_HOST, port=var.MLFLOW_PORT, concurrency=var.MLFLOW_WORKERS,
                     registry=exp_var.MLFLOW_REGISTRY_DIR.as_uri(), backend_uri=exp_var.MLFLOW_BACKEND,
                     only_ui=False, attach=False)

    elif ctx.invoked_subcommand != 'start':
        log.project_console.print('The mlflow service is not started', style='yellow')
        ctx.exit(1)


@cli.command(name='start', help='Start service')
def mlflow_start(host: str = Option(var.MLFLOW_HOST, '--host', '-h', help='Bind socket to this host.'),
                 port: int = Option(var.MLFLOW_PORT, '--port', '-p', help='Bind socket to this port.'),
                 concurrency: int = Option(var.MLFLOW_WORKERS, '-c',
                                           help='The number of mlflow server workers.'),
                 registry: Optional[str] = Option(None, '--registry', '-r',
                                                  help='Path to local directory to store artifacts.'),
                 backend_uri: Optional[str] = Option(None, '--backend', help='Backend uri.'),
                 only_ui: bool = Option(False, '--only-ui', is_flag=True, help='Launch only the Mlflow tracking UI'),
                 attach: bool = Option(False, '--attach', '-a', is_flag=True, help='Attach output and error streams')
                 ) -> None:
    """
    Start dashboard service.

    Parameters
    ----------
    host : str, default=ENV(MLFLOW_HOST) or 'localhost'
        The address where the server will listen for client and browser connections.
        Use this if you want to bind the server to a specific address. If set, the server
        will only be accessible from this address, and not from any aliases (like localhost).
    port : int, default=ENV(MLFLOW_PORT) or 8002
        The port where the server will listen for browser connections.
    concurrency : int, default=ENV(MLFLOW_WORKERS) or 1
        The number of mlflow server workers.
    registry : str, default=ENV(MLFLOW_REGISTRY_DIR) or 'file:///<BASE_DIR>/experiments/mlflow_registry'
        URI to which to persist experiment and run data. Acceptable URIs are SQLAlchemy-compatible
        database connection strings (e.g. 'sqlite:///path/to/file.db') or
        local filesystem URIs (e.g. 'file:///absolute/path/to/directory').
        By default, data will be logged to the ./mlruns directory.
    backend_uri : str, default=ENV(MLFLOW_BACKEND) or 'sqlite:///<BASE_DIR>/experiments/mlflow_registry/mlflow.db'
        Local or S3 URI to store artifacts, for new experiments. Note that this
        flag does not impact already-created experiments.
    only_ui : bool, default=False
        Launch only the Mlflow tracking UI.
    attach : bool, default=False
        Attach output and error streams.

    Raises
    ------
    typer.BadParameter:
        If concurrency option is not equal to one for windows platform.

    """

    import platform

    from trecover.config import log, exp_var
    from trecover.utils.cli import start_service

    if (is_windows := platform.system() == 'Windows') and concurrency != 1:
        raise BadParameter("Windows platform does not support concurrency option")

    command = 'ui' if only_ui else 'server'
    registry = registry or str(exp_var.MLFLOW_REGISTRY_DIR.as_uri())
    backend_uri = backend_uri or str(exp_var.MLFLOW_BACKEND)

    argv = [
        'mlflow', command,
        '--host', host,
        '--port', str(port),
        '--default-artifact-root', backend_uri,
        '--backend-store-uri', registry
    ]

    if not only_ui and not is_windows:
        argv.extend(['--workers', str(concurrency)])

    start_service(argv, name='mlflow', logfile=log.MLFLOW_LOG, pidfile=var.MLFLOW_PID)

    if attach:
        mlflow_attach(live=False)


@cli.command(name='stop', help='Stop service')
def mlflow_stop() -> None:
    """ Stop dashboard service. """

    from trecover.config import log
    from trecover.utils.cli import stop_service

    stop_service(name='mlflow', pidfile=var.MLFLOW_PID, logfile=log.MLFLOW_LOG)


@cli.command(name='status', help='Display service status')
def mlflow_status() -> None:
    """ Display dashboard service status. """

    from trecover.utils.cli import check_service

    check_service(name='mlflow', pidfile=var.MLFLOW_PID)


@cli.command(name='attach', help='Attach local output stream to a service')
def mlflow_attach(live: bool = Option(False, '--live', '-l', is_flag=True,
                                      help='Stream only fresh log records')
                  ) -> None:
    """
    Attach local output stream to a running dashboard service.

    Parameters
    ----------
    live : bool, Default=False
        Stream only fresh log records.

    """

    from trecover.config import log
    from trecover.utils.cli import stream

    with log.project_console.screen():
        for record in stream(('mlflow', log.MLFLOW_LOG), live=live):
            log.project_console.print(record)

    log.project_console.clear()


if __name__ == '__main__':
    cli()



from typer import Typer, Option, Argument, Context

from trecover.config import var

cli = Typer(name='API-cli', add_completion=False, help='Manage API service')


@cli.callback(invoke_without_command=True)
def api_state_verification(ctx: Context) -> None:
    """
    Perform cli commands verification (state checking).

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

    """

    from trecover.config import log

    if var.API_PID.exists():
        if ctx.invoked_subcommand in ('start', None):
            log.project_console.print(':rocket: The API service is already started', style='bright_blue')
            ctx.exit(0)

    elif ctx.invoked_subcommand is None:
        api_start(host=var.FASTAPI_HOST, port=var.FASTAPI_PORT, loglevel=var.LogLevel.info,
                  concurrency=var.FASTAPI_WORKERS, attach=False, no_daemon=False)

    elif ctx.invoked_subcommand not in ('start', 'params', 'recover'):
        log.project_console.print('The API service is not started', style='yellow')
        ctx.exit(1)


@cli.command(name='config', help='Receive configuration or specific parameter of the model used for inference')
def api_config(url: str = Option(var.FASTAPI_URL, help='API url'),
               param: str = Option(None, help='Param name to receive')
               ) -> None:
    """
    Receive configuration or specific parameter of the model used for inference.

    Parameters
    ----------
    url : str, default=ENV(FASTAPI_URL) or 'http://localhost:8001'
        API url.
    param : str, default=None
        Param name whose value to receive. Receive all configuration values if None.

    """

    import json
    import requests

    from trecover.config import log

    if param:
        response = requests.get(url=f'{url}/config/{param}')
    else:
        response = requests.get(url=f'{url}/config')

    log.project_console.print(json.dumps(response.json(), indent=4))


@cli.command(name='recover', help='Send keyless reading API request')
def api_recover(data_path: str = Argument(..., help='Path to file or dir for data'),
                url: str = Option(var.FASTAPI_URL, help='API url'),
                separator: str = Option(' ', help='Columns separator in the input files'),
                noisy: bool = Option(False, help='Input files are noisy texts'),
                min_noise: int = Option(3, help='Min noise parameter. Minimum value is alphabet size'),
                max_noise: int = Option(5, help='Max noise parameter. Maximum value is alphabet size'),
                beam_width: int = Option(1, help='Width for beam search algorithm. Maximum value is alphabet size'),
                n_to_show: int = Option(0, help='Number of columns to visualize. Zero value means for no restrictions'),
                delimiter: str = Option('', help='Delimiter for columns visualization')
                ) -> None:
    """
    Send keyless reading API request.

    Parameters
    ----------
    data_path : Path
        Path to file or dir for data.
    url : str
        API url.
    separator : str, default=' '
        Columns separator in the input files.
    noisy : bool, default=False
        Indicates that input files are noisy texts.
    min_noise : int, default=3
        Min noise size per column. Minimum value is zero.
    max_noise : int, default=5
        Max noise size per column. Maximum value is alphabet size.
    beam_width : int, default=5
        Width for beam search algorithm. Maximum value is alphabet size.
    n_to_show : int, default=0
        Number of columns to visualize. Zero value means for no restriction's.
    delimiter : str, default=''
        Delimiter for columns visualization.

    Examples
    --------
    >>> trecover api recover examples/example_1.txt
    ╭──────────────────────────────────────────────────── example_1.txt ───────────────────────────────────────────────╮
    │                                                        Columns                                                   │
    │ ajocmbfeafodadbddciafqnahdfeihhkieeaacacafkdchddakhecmmlibfinaehbcbdiicejkeahnfemaeaadbkagacbdmahbibacfddfbbbca… │
    │ enpenkhgglrifflheioentrmjenkjnrmlhphdddeihliekeeeolflonpmctjolgkdeljjmljmmjiisjknjghgeelhkbddlpjjekrkdkilgiocii… │
    │ gsxtoplqkrtknksinktipwvnlnqqrstotoqspoejtsnoiuoflpohvtovqeutunjojlmksonosskpvxporrltnfgoprdemstnshnssgnronjreqj… │
    │ xvzwttqtxvxuoptowuxnxyzrwrrtwtyqwqvutrwrxvtxxwurrtqlwuqzvnwvxossmmpnutosuxlswyuvtttvqulrqzrrwuxtyqouwiuupwsxnrm… │
    │  y y yz zy  y w zy uz  yys   u tzs   x u         wx     wy w tuvpuwu  x yyowyz  z  wxyu     xyy   v yr    t yvw… │
    │                                                       Predicted                                                  │
    │ enpeoplearoundthecountrywereintothestreetstickedatheconvictionsspewditnessesinpentlandboardeddytheirwindowsbyra… │
    │                                                                                                                  │
    ╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯
    Elapsed:   4.716 s

    Notes
    -----
    A larger "beam_width" parameter value can improve keyless reading, but it will also take longer to compute.

    """

    import requests
    from http import HTTPStatus
    from time import time, sleep
    from pathlib import Path

    from rich.console import Group
    from rich.panel import Panel
    from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn, TimeElapsedColumn
    from rich.text import Text

    from trecover.config import log
    from trecover.utils.cli import get_files_columns
    from trecover.utils.visualization import visualize_columns

    data_path = Path(data_path).absolute()

    if not noisy and min_noise >= max_noise:
        log.project_logger.error('[red]Maximum noise range must be grater than minimum noise range')
        return

    if not any([data_path.is_file(), data_path.is_dir()]):
        log.project_logger.error('[red]Files for inference needed to be specified')
        return

    files, files_columns = get_files_columns(data_path, separator, noisy, min_noise, max_noise, n_to_show)
    payload = {
        'columns': None,
        'beam_width': beam_width,
        'delimiter': delimiter
    }

    for file_id, (file, file_columns) in enumerate(zip(files, files_columns), start=1):
        start_time = time()

        file_columns = [''.join(set(c)) for c in file_columns]

        payload['columns'] = file_columns
        task_info = requests.post(url=f'{url}/recover', json=payload)
        task_info = task_info.json()

        if not task_info['task_id']:
            log.project_logger.error(f'{file_id}/{len(files_columns)} [red]Failed {file.name}:\n'
                                     f'{task_info["message"]}')
            continue

        task_status = requests.get(url=f'{url}/status/{task_info["task_id"]}')
        task_status = task_status.json()

        label = f'{file_id}/{len(files_columns)} Processing {file.name}'

        with Progress(
                TextColumn('{task.description}', style='bright_blue'),
                BarColumn(complete_style='bright_blue'),
                TextColumn('{task.percentage:>3.0f}%', style='bright_blue'),
                TextColumn('Remaining', style='bright_blue'),
                TimeRemainingColumn(),
                TextColumn('Elapsed', style='bright_blue'),
                TimeElapsedColumn(),
                transient=True,
        ) as progress:
            request_progress = progress.add_task(label, total=len(file_columns))

            while task_status['status_code'] == HTTPStatus.PROCESSING:
                task_status = requests.get(url=f'{url}/status/{task_info["task_id"]}')
                task_status = task_status.json()

                progress.update(request_progress, completed=task_status['progress'])

                sleep(0.5)

        requests.delete(url=f'{url}/{task_info["task_id"]}')

        if task_status['status_code'] != HTTPStatus.OK:
            log.project_logger.error(f'{file_id}/{len(files_columns)} [red]Failed {file.name}:\n'
                                     f'{task_status["message"]}')
            continue

        columns = visualize_columns(file_columns, delimiter=delimiter, as_rows=True)
        columns = (Text(row, style='bright_blue', overflow='ellipsis', no_wrap=True) for row in columns)

        chains = [Text(chain, style='cyan', justify='center', overflow='ellipsis', end='\n\n')
                  for (chain, _) in task_status['chains']]

        panel_group = Group(
            Text('Columns', style='magenta', justify='center'),
            *columns,
            Text('Predicted', style='magenta', justify='center'),
            *chains
        )

        log.project_console.print(
            Panel(panel_group, title=file.name, border_style='magenta'),
            justify='center'
        )

        log.project_console.print(f'\nElapsed: {time() - start_time:>7.3f} s\n', style='bright_blue')


@cli.command(name='start', help='Start service')
def api_start(host: str = Option(var.FASTAPI_HOST, '--host', '-h', help='Bind socket to this host.'),
              port: int = Option(var.FASTAPI_PORT, '--port', '-p', help='Bind socket to this port.'),
              loglevel: var.LogLevel = Option(var.LogLevel.info, '--loglevel', '-l', help='Logging level.'),
              concurrency: int = Option(var.FASTAPI_WORKERS, '-c', help='The number of worker processes.'),
              attach: bool = Option(False, '--attach', '-a', is_flag=True,
                                    help='Attach output and error streams'),
              no_daemon: bool = Option(False, '--no-daemon', is_flag=True, help='Do not run as a daemon process')
              ) -> None:
    """
    Start API service.

    Parameters
    ----------
    host : str, default=ENV(FASTAPI_HOST) or 'localhost'
        Bind socket to this host.
    port : int, default=ENV(FASTAPI_PORT) or 8001
        Bind socket to this port.
    loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.
    concurrency : int, default=ENV(FASTAPI_WORKERS) or 1
        The number of worker processes.
    attach : bool, default=False
        Attach output and error streams.
    no_daemon : bool, default=False
        Do not run as a daemon process.

    """

    from subprocess import run

    from trecover.config import log
    from trecover.utils.cli import start_service

    argv = [
        'uvicorn', 'trecover.app.api.trecoverapi:api',
        '--host', host,
        '--port', str(port),
        '--workers', str(concurrency),
        '--log-level', loglevel
    ]

    if no_daemon:
        run(argv)
    else:
        start_service(argv, name='API', logfile=log.API_LOG, pidfile=var.API_PID)

        if attach:
            api_attach(live=False)


@cli.command(name='stop', help='Stop service')
def api_stop() -> None:
    """ Stop API service. """

    from trecover.config import log
    from trecover.utils.cli import stop_service

    stop_service(name='API', pidfile=var.API_PID, logfile=log.API_LOG)


@cli.command(name='status', help='Display service status')
def api_status() -> None:
    """ Display API service status. """

    from trecover.utils.cli import check_service

    check_service(name='API', pidfile=var.API_PID)


@cli.command(name='attach', help='Attach local output stream to a service')
def api_attach(live: bool = Option(False, '--live', '-l', is_flag=True,
                                   help='Stream only fresh log records')
               ) -> None:
    """
    Attach local output stream to a running API service.

    Parameters
    ----------
    live : bool, Default=False
        Stream only fresh log records

    """

    from trecover.config import log
    from trecover.utils.cli import stream

    with log.project_console.screen():
        for record in stream(('API', log.API_LOG), live=live):
            log.project_console.print(record)

    log.project_console.clear()


if __name__ == '__main__':
    cli()



import json
from pathlib import Path
from typing import Dict, Optional

import torch
from rich.prompt import Confirm

from trecover.config import log
from trecover.model import TRecover
from trecover.utils.train import ExperimentParams


def get_recent_weights_path(exp_dir: Path,
                            exp_mark: str,
                            weights_name: Optional[str] = None
                            ) -> Optional[Path]:
    """
    Get a model recent weights path.

    Parameters
    ----------
    exp_dir : Path
        Experiment directory path.
    exp_mark : str
        Experiment folder mark.
    weights_name : str, default=None
        Weights filename.

    Returns
    -------
    Optional[Path]:
        Recent weights path if it exists otherwise None object.

    """

    if weights_name:
        return weights_path if (weights_path := exp_dir / exp_mark / weights_name).exists() else None

    if (weights_path := exp_dir / exp_mark / 'weights').exists():
        recent_weights = None
        most_recent_timestamp = 0

        for weights in weights_path.iterdir():
            if timestamp := weights.stat().st_ctime > most_recent_timestamp:
                recent_weights = weights
                most_recent_timestamp = timestamp

        return recent_weights


def get_model(token_size: int,
              pe_max_len: int,
              num_layers: int,
              d_model: int,
              n_heads: int,
              d_ff: int,
              dropout: float,
              device: torch.device = torch.device('cpu'),
              weights: Optional[Path] = None,
              silently: bool = False
              ) -> TRecover:
    """
    Get a model with specified configuration.

    Parameters
    ----------
    token_size : int
        Token (column) size.
    pe_max_len : int
        Positional encoding max length.
    num_layers : int
        Number of encoder and decoder blocks
    d_model : int
        Model dimension - number of expected features in the encoder (decoder) input.
    n_heads : int
        Number of encoder and decoder attention heads.
    d_ff : int
        Dimension of the feedforward layer.
    dropout : float,
        Dropout range.
    device : torch.device, default=torch.device('cpu')
        Device on which to allocate the model.
    weights : Path, default=None
        Model weights path for initialization.
    silently : bool, default=False
        Initialize the model silently without any verbose information.

    Returns
    -------
    model : TRecover
        Initialized model.

    Raises
    ------
    SystemExit:
        If the weight's path is not provided and the cli 'stop' option is selected.

    """

    model = TRecover(token_size, pe_max_len, num_layers, d_model, n_heads, d_ff, dropout).to(device)

    if weights and weights.exists() and weights.is_file():
        model.load_parameters(weights, device=device)

        if not silently:
            log.project_console.print(f'The below model parameters have been loaded:\n{weights}',
                                      style='bright_green')
        return model

    if weights:
        log.project_console.print(f'Failed to load model parameters: {str(weights)}', style='bold red')
    else:
        log.project_console.print("Model parameters aren't specified", style='bright_blue')

    if silently or Confirm.ask(prompt='[bright_blue]Continue training from scratch?', default=True,
                               console=log.project_console):
        return model
    else:
        raise SystemExit


def load_params(model_params: Path) -> ExperimentParams:
    """
    Get experiment parameters container.

    Parameters
    ----------
    model_params : Path
        Path to serialized experiment parameters.

    Returns
    -------
    ExperimentParams:
        Experiment parameters container as a ExperimentParams object.

    """

    return ExperimentParams(json.load(model_params.open()))


def save_params(data: Dict, filepath: Path, sort=False) -> None:
    """
    Save experiment parameters on disk.

    Parameters
    ----------
    data : Dict
        Experiment parameters.
    filepath : Path
        File path for saving.
    sort : bool, default=False
        Perform parameters keys sorting.

    """

    with filepath.open('w') as f:
        json.dump(data, indent=2, fp=f, sort_keys=sort)



from celery import Celery
from celery.signals import after_setup_task_logger, after_setup_logger

from trecover.config import var, log

celery_app = Celery('TRecover',
                    broker=var.CELERY_BROKER,
                    backend=var.CELERY_BACKEND,
                    include=['trecover.app.api.backend.tasks'])

celery_app.conf.update({
    'task_serializer': 'json',
    'result_serializer': 'json',
    'accept_content': ['json'],
    'worker_prefetch_multiplier': 1,
    'task_acks_late': True,
    'task_track_started': True,
    'task_reject_on_worker_lost': True,
})


@after_setup_task_logger.connect
def setup_task_logger(logger, *args, **kwargs):
    """ Customize celery task logger. """

    logger.addHandler(hdlr=log.error_handler)
    logger.addHandler(hdlr=log.info_handler)


@after_setup_logger.connect
def setup_task_logger(logger, *args, **kwargs):
    """ Customize celery logger. """

    logger.addHandler(hdlr=log.error_handler)
    logger.addHandler(hdlr=log.info_handler)



from typing import List, Optional

import docker
from docker import DockerClient
from docker.models.containers import Container
from docker.models.images import Image
from docker.models.volumes import Volume
from rich.live import Live

from trecover.config import log


def get_client() -> DockerClient:
    """ Get the docker client """

    return docker.from_env()


def is_docker_running() -> bool:
    """ Check if docker is running """
    try:
        get_client().ping()
    except (docker.errors.APIError, docker.errors.DockerException):
        return False

    return True


def get_images_list() -> List[str]:
    """ Get list of docker images """
    image_names = list()

    for image in get_client().images.list():
        image_names.extend(image.tags)

    return image_names


def get_containers_list() -> List[str]:
    """ Get list of docker containers """

    return [container.name for container in get_client().containers.list(all=True)]


def get_container(id_or_name: str) -> Optional[Container]:
    """
    Get the docker container with given name or id.

    Parameters
    ----------
    id_or_name : str
        Docker container id or name.

    Returns
    -------
    Optional[Container]:
        Container instance if it exists, otherwise None.

    """

    try:
        return get_client().containers.get(id_or_name)
    except docker.errors.NotFound:
        return None


def get_volume(id_or_name: str) -> Optional[Volume]:
    """
    Get the docker volume with given name or id.

    Parameters
    ----------
    id_or_name : str
        Volume id or name.

    Returns
    -------
    Optional[Volume]:
        Volume instance if it exists, otherwise None.

    """

    try:
        return get_client().volumes.get(id_or_name)
    except docker.errors.NotFound:
        return None


def get_image(name: str) -> Optional[Image]:
    """
    Get the docker image with given name.

    Parameters
    ----------
    name : str
        Docker image name.

    Returns
    -------
    Optional[Image]:
        Image instance if it exists, otherwise None.

    """

    try:
        return get_client().images.get(name)
    except docker.errors.NotFound:
        return None


def pull_image(name: str) -> Image:
    """
    Pull (download) the docker image with given name.

    Parameters
    ----------
    name : str
        Docker image name.

    Returns
    -------
    Image:
        Image instance of the pulled docker image.

    """

    with Live(console=log.project_console) as screen:
        screen.update('[bright_blue]Waiting')
        for state in get_client().api.pull(name, stream=True, decode=True):
            try:
                screen.update(f'[bright_blue]{state["status"]}[/] {state["progress"]}')
            except KeyError:
                screen.update(f'[bright_blue]{state["status"]}')

    return get_image(name)



from typing import Tuple, List, Dict

from trecover.app.api.backend.celeryapp import celery_app
from trecover.app.api.backend.tasksbase import ModelConfigTask, PredictTask


@celery_app.task(bind=True, base=ModelConfigTask)
def get_model_config(self: ModelConfigTask) -> Dict:
    """
    Celery task implementation that returns values of the model configuration.

    Parameters
    ----------
    self : ModelConfigTask
        Celery task base class.

    Returns
    -------
    self.config : Dict
        The values of the model configuration.

    """

    return self.config


@celery_app.task(bind=True, base=PredictTask)
def predict(self: PredictTask,
            columns: List[str],
            beam_width: int,
            delimiter: str
            ) -> Tuple[List[str], List[Tuple[str, float]]]:
    """
    Celery task implementation that performs keyless reading.

    Parameters
    ----------
    self : ArtifactsTask
        Celery task base class.
    columns : List[str]
        Columns to keyless read.
    beam_width : int
        Width for beam search algorithm. Maximum value is alphabet size.
    delimiter : str
        Delimiter for columns visualization.

    Returns
    -------
    (columns, chains) : Tuple[List[str], List[Tuple[str, float]]]
        The columns and read chains.

    Raises
    ------
    AssertionError:
        If the number of columns is grater than self.model.pe_max_len.

    """

    from trecover.utils.beam_search import beam_search, celery_task_loop
    from trecover.utils.transform import columns_to_tensor, tensor_to_target
    from trecover.utils.visualization import visualize_target

    assert len(columns) <= self.model.pe_max_len, f'Number of columns must be less than {self.model.pe_max_len}.'

    src = columns_to_tensor(columns, self.device)

    chains = beam_search(src, self.model, beam_width, self.device, beam_loop=celery_task_loop(self))
    chains = [(visualize_target(tensor_to_target(chain), delimiter=delimiter), prob) for (chain, prob) in chains]

    return columns, chains



import os

from trecover.config import var

TRAIN_DATA = var.DATA_DIR / 'train'
VAL_DATA = var.DATA_DIR / 'validation'
VIS_DATA = var.DATA_DIR / 'visualization'

var.DATA_DIR.mkdir(parents=True, exist_ok=True)
TRAIN_DATA.mkdir(parents=True, exist_ok=True)
VAL_DATA.mkdir(parents=True, exist_ok=True)
VIS_DATA.mkdir(parents=True, exist_ok=True)

WANDB_REGISTRY_DIR = var.EXPERIMENTS_DIR / 'wandb_registry'
MLFLOW_REGISTRY_DIR = var.EXPERIMENTS_DIR / 'mlflow_registry'
MLFLOW_BACKEND = os.getenv('MLFLOW_BACKEND',
                           default=f'sqlite:{(MLFLOW_REGISTRY_DIR.absolute() / "mlflow.db").as_uri()[5:]}')

var.EXPERIMENTS_DIR.mkdir(parents=True, exist_ok=True)
MLFLOW_REGISTRY_DIR.mkdir(parents=True, exist_ok=True)
WANDB_REGISTRY_DIR.mkdir(parents=True, exist_ok=True)



import uuid
from pathlib import Path
from typing import Dict, Union, Optional, Any

import mlflow
import wandb
from mlflow.tracking.fluent import ActiveRun, end_run
from wandb.wandb_run import Run

from trecover.config import exp_var
from trecover.utils.train import get_experiment_mark


class BaseMonitor(object):
    # TODO docs

    def log_metrics(self, metrics: Dict[str, Union[float, int]], step: Optional[int] = None) -> None:
        # TODO docs

        raise NotImplementedError

    def log_artifact(self, path: str) -> None:
        # TODO docs

        raise NotImplementedError

    def log_variables(self, variables: Dict[str, Union[float, int]]) -> None:
        # TODO docs

        raise NotImplementedError

    def start(self):
        # TODO docs

        raise NotImplementedError

    def finish(self):
        # TODO docs

        raise NotImplementedError

    def __enter__(self) -> Union[Run, ActiveRun]:
        # TODO docs

        return self.start(),

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # TODO docs

        return self.finish()


class IdentityMonitor(BaseMonitor):
    # TODO docs

    def __init__(self,
                 project_name: str = None,
                 experiment_name: str = None,
                 config: Dict[str, Any] = None):
        self.project_name = project_name or f'DefaultProject-{uuid.uuid4().hex}'
        self.experiment_name = experiment_name or get_experiment_mark()
        self.config = config or dict()

    def log_metrics(self, metrics: Dict[str, Union[float, int]], step: Optional[int] = None) -> None:
        pass

    def log_artifact(self, path: str) -> None:
        pass

    def log_variables(self, variables: Dict[str, Union[float, int]]) -> None:
        pass

    def start(self):
        pass

    def finish(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class WandbMonitor(BaseMonitor):
    # TODO docs

    def __init__(self,
                 project_name: str,
                 experiment_name: str,
                 config: Dict[str, Any],
                 registry: Union[str, Path] = exp_var.WANDB_REGISTRY_DIR.absolute()):
        self.project_name = project_name
        self.experiment_name = experiment_name
        self.config = config
        self.registry = registry

    def log_metrics(self, metrics: Dict[str, Union[float, int]], step: Optional[int] = None) -> None:
        wandb.log(metrics)

    def log_artifact(self, path: str) -> None:
        wandb.log_artifact(artifact_or_path=path, type='experiment_artifact')

    def log_variables(self, variables: Dict[str, Union[float, int]]) -> None:
        for var_name, var_value in variables.items():
            wandb.summary[var_name] = var_value

    def start(self) -> Run:
        return wandb.init(project=self.project_name,
                          name=self.experiment_name,
                          config=self.config,
                          dir=self.registry)

    def finish(self) -> None:
        wandb.finish()


class MlflowMonitor(BaseMonitor):
    # TODO docs

    def __init__(self,
                 project_name: str,
                 experiment_name: str,
                 config: Dict[str, Any],
                 tracking_uri: Union[str, Path] = exp_var.MLFLOW_REGISTRY_DIR.absolute().as_uri()):
        self.project_name = project_name
        self.experiment_name = experiment_name
        self.config = config
        self.tracking_uri = tracking_uri

    def log_metrics(self, metrics: Dict[str, Union[float, int]], step: Optional[int] = None) -> None:
        mlflow.log_metrics(metrics, step=step)

    def log_artifact(self, path: str) -> None:
        mlflow.log_artifact(path)

    def log_variables(self, variables: Dict[str, Union[float, int]]) -> None:
        mlflow.log_params(variables)

    def start(self) -> ActiveRun:
        mlflow.set_tracking_uri(self.tracking_uri)
        mlflow.set_experiment(self.project_name)
        mlflow.log_params(self.config)

        with mlflow.start_run(run_name=self.experiment_name, nested=True) as run:
            return run

    def finish(self):
        end_run()



from typer import Typer, Option, Context

from trecover.config import var, log

cli = Typer(name='Dashboard-cli', add_completion=False, help='Manage Dashboard service')


@cli.callback(invoke_without_command=True)
def dashboard_state_verification(ctx: Context) -> None:
    """
    Perform cli commands verification (state checking).

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

    """

    from trecover.config import log

    if var.DASHBOARD_PID.exists():
        if ctx.invoked_subcommand in ('start', None):
            log.project_console.print(':rocket: The dashboard service is already started', style='bright_blue')
            ctx.exit(0)

    elif ctx.invoked_subcommand is None:
        dashboard_start(host=var.STREAMLIT_HOST, port=var.STREAMLIT_PORT, loglevel=var.LogLevel.info, attach=False,
                        no_daemon=False)

    elif ctx.invoked_subcommand != 'start':
        log.project_console.print('The dashboard service is not started', style='yellow')
        ctx.exit(1)


@cli.command(name='start', help='Start service')
def dashboard_start(host: str = Option(var.STREAMLIT_HOST, '--host', '-h', help='Bind socket to this host.'),
                    port: int = Option(var.STREAMLIT_PORT, '--port', '-p', help='Bind socket to this port.'),
                    loglevel: var.LogLevel = Option(var.LogLevel.info, '--loglevel', '-l', help='Logging level.'),
                    attach: bool = Option(False, '--attach', '-a', is_flag=True,
                                          help='Attach output and error streams'),
                    no_daemon: bool = Option(False, '--no-daemon', is_flag=True, help='Do not run as a daemon process')
                    ) -> None:
    """
    Start dashboard service.

    Parameters
    ----------
    host : str, default=ENV(STREAMLIT_HOST) or 'localhost'
        The address where the server will listen for client and browser connections.
        Use this if you want to bind the server to a specific address. If set, the server
        will only be accessible from this address, and not from any aliases (like localhost).
    port : int, default=ENV(STREAMLIT_PORT) or 8000
        The port where the server will listen for browser connections.
    loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.
    attach : bool, default=False
        Attach output and error streams.
    no_daemon : bool, default=False
        Do not run as a daemon process.

    """

    from subprocess import run

    from trecover.config import log
    from trecover.app import dashboard
    from trecover.utils.cli import start_service

    argv = ['streamlit',
            'run', dashboard.__file__,
            '--server.address', host,
            '--server.port', str(port),
            '--logger.level', loglevel,
            '--global.suppressDeprecationWarnings', 'True',
            '--theme.backgroundColor', '#E7EAD9',
            '--theme.secondaryBackgroundColor', '#DFE3D0',
            '--theme.primaryColor', '#FF8068',
            '--theme.textColor', '#157D96'
            ]

    if no_daemon:
        run(argv)
    else:
        start_service(argv, name='dashboard', logfile=log.DASHBOARD_LOG, pidfile=var.DASHBOARD_PID)

        if attach:
            dashboard_attach(live=False)


@cli.command(name='stop', help='Stop service')
def dashboard_stop() -> None:
    """ Stop dashboard service. """

    from trecover.utils.cli import stop_service

    stop_service(name='dashboard', pidfile=var.DASHBOARD_PID, logfile=log.DASHBOARD_LOG)


@cli.command(name='status', help='Display service status')
def dashboard_status() -> None:
    """ Display dashboard service status. """

    from trecover.utils.cli import check_service

    check_service(name='dashboard', pidfile=var.DASHBOARD_PID)


@cli.command(name='attach', help='Attach local output stream to a service')
def dashboard_attach(live: bool = Option(False, '--live', '-l', is_flag=True,
                                         help='Stream only fresh log records')
                     ) -> None:
    """
    Attach local output stream to a running dashboard service.

    Parameters
    ----------
    live : bool, Default=False
        Stream only fresh log records

    """

    from trecover.config import log
    from trecover.utils.cli import stream

    with log.project_console.screen():
        for record in stream(('dashboard', log.DASHBOARD_LOG), live=live):
            log.project_console.print(record)

    log.project_console.clear()


if __name__ == '__main__':
    cli()



from typer import Typer, Option, Context, BadParameter

from trecover.config import var

cli = Typer(name='Worker-cli', add_completion=False, help='Manage Worker service')


@cli.callback(invoke_without_command=True)
def worker_state_verification(ctx: Context) -> None:
    """
    Perform cli commands verification (state checking).

    Parameters
    ----------
    ctx : Context
        Typer (Click like) special internal object that holds state relevant
        for the script execution at every single level.

   """

    from trecover.config import log

    if var.WORKER_PID.exists():
        if ctx.invoked_subcommand in ('start', None):
            log.project_console.print(':rocket: The worker service is already started', style='bright_blue')
            ctx.exit(0)

    elif ctx.invoked_subcommand is None:
        worker_start(name='TRecoverWorker', pool=var.PoolType.solo, loglevel=var.LogLevel.info,
                     concurrency=var.CELERY_WORKERS, broker_url=var.CELERY_BROKER,
                     backend_url=var.CELERY_BACKEND, attach=False, no_daemon=False)

    elif ctx.invoked_subcommand != 'start':
        log.project_console.print('The worker service is not started', style='yellow')
        ctx.exit(1)


@cli.command(name='start', help='Start service')
def worker_start(name: str = Option('TRecoverWorker', '--name', '-n', help='Set custom worker hostname.'),
                 pool: var.PoolType = Option(var.PoolType.solo, '--pool', '-p',
                                             help='Worker processes/threads pool type.'),
                 loglevel: var.LogLevel = Option(var.LogLevel.info, '--loglevel', '-l', help='Logging level.'),
                 concurrency: int = Option(var.CELERY_WORKERS, '-c', help='The number of worker processes.'),
                 broker_url: str = Option(var.CELERY_BROKER, '--broker', help='Broker url.'),
                 backend_url: str = Option(var.CELERY_BACKEND, '--backend', help='Backend url.'),
                 attach: bool = Option(False, '--attach', '-a', is_flag=True, help='Attach output and error streams'),
                 no_daemon: bool = Option(False, '--no-daemon', is_flag=True, help='Do not run as a daemon process')
                 ) -> None:
    """
    Start API service.

    Parameters
    ----------
    name : str, default='TRecoverWorker'
        Custom worker hostname.
    pool : str, {'prefork', 'eventlet', 'gevent', 'processes', 'solo'}, default='solo'
        Worker processes/threads pool type.
    loglevel : {'debug', 'info', 'warning', 'error', 'critical'}, default='info'
        Level of logging.
    concurrency : int, default=ENV(CELERY_WORKERS) or 1
        The number of worker processes.
    broker_url : str, default=ENV(CELERY_BROKER) or 'pyamqp://guest@localhost'
        Broker url.
    backend_url : str, default=ENV(CELERY_BACKEND) or 'redis://localhost'
        Backend url.
    attach : bool, default=False
        Attach output and error streams.
    no_daemon : bool, default=False
        Do not run as a daemon process.

    Raises
    ------
    typer.BadParameter:
        If non-solo pool type is selected for windows platform.

    """

    import platform
    from subprocess import run

    from trecover.config import log
    from trecover.utils.cli import start_service

    if platform.system() == 'Windows' and pool != var.PoolType.solo:
        raise BadParameter("Windows platform only supports 'solo' pool")

    argv = [
        'celery',
        '--broker', broker_url,
        '--result-backend', backend_url,
        '--app', 'trecover.app.api.backend.celeryapp', 'worker',
        '--hostname', name,
        '--concurrency', str(concurrency),
        '--pool', pool,
        '--loglevel', loglevel
    ]

    if no_daemon:
        run(argv)
    else:
        start_service(argv, name='worker', logfile=log.WORKER_LOG, pidfile=var.WORKER_PID)

        if attach:
            worker_attach(live=False)


@cli.command(name='stop', help='Stop service')
def worker_stop() -> None:
    """ Stop worker service. """

    from trecover.config import log
    from trecover.utils.cli import stop_service

    stop_service(name='worker', pidfile=var.WORKER_PID, logfile=log.WORKER_LOG)


@cli.command(name='status', help='Display service status')
def worker_status() -> None:
    """ Display worker service status. """

    from trecover.utils.cli import check_service

    check_service(name='worker', pidfile=var.WORKER_PID)


@cli.command(name='attach', help='Attach local output stream to a service')
def worker_attach(live: bool = Option(False, '--live', '-l', is_flag=True,
                                      help='Stream only fresh log records')
                  ) -> None:
    """
    Attach local output stream to a running worker service.

    Parameters
    ----------
    live : bool, Default=False
        Stream only fresh log records.

    """

    from trecover.config import log
    from trecover.utils.cli import stream

    with log.project_console.screen():
        for record in stream(('worker', log.WORKER_LOG), live=live):
            log.project_console.print(record)

    log.project_console.clear()


if __name__ == '__main__':
    cli()



import platform
import time
from http import HTTPStatus
from typing import Tuple, List

import requests
import streamlit as st
from requests.exceptions import ConnectionError

from trecover.config import var
from trecover.utils.inference import data_to_columns, create_noisy_columns
from trecover.utils.visualization import visualize_columns


def main() -> None:
    st.set_page_config(
        page_title='TRecover',
        page_icon='🩹',
        layout='wide',
        initial_sidebar_state='expanded')

    if 'history' not in st.session_state:
        st.session_state.history = list()

    if 'data' not in st.session_state:
        st.session_state.data = ''

    if 'regenerate' not in st.session_state:
        st.session_state.regenerate = False

    if 'stop' not in st.session_state:
        st.session_state.stop = False

    if 'columns' not in st.session_state:
        st.session_state.columns = None

    if 'task_id' not in st.session_state:
        st.session_state.task_id = None

    if 'is_unix' not in st.session_state:
        st.session_state.is_unix = platform.system() != 'Windows'

    sidebar()


def set_regenerate() -> None:
    st.session_state.regenerate = True


def unset_regenerate() -> None:
    st.session_state.regenerate = False


def set_stop() -> None:
    st.session_state.stop = True


def unset_stop() -> None:
    st.session_state.stop = False


def sidebar() -> None:
    st.sidebar.header('Sections')

    option = st.sidebar.radio('', ('Inference', 'Inference history'))

    if option == 'Inference':
        is_plain, min_noise, max_noise, bw = inference_sidebar()
        inference_page(is_plain, min_noise, max_noise, bw)
    else:
        history_sidebar()
        history_page()


def inference_sidebar() -> Tuple[bool, int, int, int]:
    st.sidebar.text('\n')

    st.sidebar.header('Data parameters')

    data_type = st.sidebar.radio('Input type', ('Plain text', 'Noisy columns'), key='data_type',
                                 index=0 if 'Plain text' == st.session_state.get('data_type', 'Plain text') else 1)
    is_plain = data_type == 'Plain text'

    st.sidebar.text('\n')

    if is_plain:
        min_noise, max_noise = st.sidebar.slider('\nNoise range', 0, 25, key='noise_range',
                                                 value=st.session_state.get('noise_range', (0, 5)),
                                                 on_change=set_regenerate)
    else:
        min_noise, max_noise = 0, 0

    bw = st.sidebar.slider('Beam search width', 1, 26, key='beam_width',
                           value=st.session_state.get('beam_width', 5))

    if max_noise > var.MAX_NOISE:
        st.sidebar.warning('Max noise value is too large. This will entail poor performance')

    return is_plain, min_noise, max_noise + 1, bw


def history_sidebar() -> None:
    pass


def save_to_history(is_plain: bool,
                    min_noise: int,
                    max_noise: int,
                    bw: int,
                    columns: List[str],
                    chains: List[Tuple[str, float]]
                    ) -> None:
    text = st.session_state.data if is_plain else None

    st.session_state.history.append((is_plain, text, min_noise, max_noise, bw, columns, chains))


@st.cache(ttl=3600, show_spinner=False, suppress_st_warning=True)
def predict(columns: List[str], bw: int) -> List[Tuple[str, float]]:
    try:
        payload = {
            'columns': columns,
            'beam_width': bw
        }

        if not st.session_state.task_id:
            task_info = requests.post(url=f'{var.FASTAPI_URL}/recover', json=payload)
            task_info = task_info.json()

            if not task_info.get('task_id'):
                st.error(task_info)
                st.stop()

            st.session_state.task_id = task_info["task_id"]

        st.info('Requesting')

        status = requests.get(url=f'{var.FASTAPI_URL}/status/{st.session_state.task_id}')
        status = status.json()

        st.info('Requested')

        while status['status_code'] == HTTPStatus.PROCESSING and status['state'] != 'PREDICT':
            if status['state'] == 'LOADING':
                st.info('Wait a moment: Model is loading')
                time.sleep(0.3)
            elif status['state'] == 'PENDING':
                st.info('Wait a moment: Task will be executed as soon as the celery worker is free')
                time.sleep(1)
            else:
                time.sleep(0.1)

            status = requests.get(url=f'{var.FASTAPI_URL}/status/{st.session_state.task_id}')
            status = status.json()

        progress_bar = st.progress(0)

        while status['status_code'] == HTTPStatus.PROCESSING:
            status = requests.get(url=f'{var.FASTAPI_URL}/status/{st.session_state.task_id}')
            status = status.json()

            completed = status['progress'] or 0
            progress_bar.progress(completed / len(columns))

            time.sleep(0.3)

        requests.delete(url=f'{var.FASTAPI_URL}/{st.session_state.task_id}')

        if status['status_code'] != HTTPStatus.OK:
            st.error(status['message'])
            st.stop()

        st.session_state.task_id = None

        return status['chains']

    except ConnectionError:
        st.error(f'It seems that the API service is not running.\n\n'
                 f'Failed to establish a {var.FASTAPI_URL} connection.')
        st.stop()


def stop_prediction() -> None:
    try:
        if st.session_state.task_id:
            requests.delete(url=f'{var.FASTAPI_URL}/{st.session_state.task_id}')
            st.session_state.task_id = None

        unset_stop()

    except ConnectionError:
        st.error(f'It seems that the API service is not running.\n\n'
                 f'Failed to establish a {var.FASTAPI_URL} connection.')
        st.stop()


def get_noisy_columns(data: str, min_noise: int, max_noise: int) -> List[str]:
    columns = create_noisy_columns(data, min_noise, max_noise)

    return [''.join(set(c)) for c in columns]  # kinda shuffle columns


def inference_page(is_plain: bool, min_noise: int, max_noise: int, bw: int) -> None:
    input_label = 'Insert plain data' if is_plain else 'Insert noisy columns separated by spaces'
    st.subheader(input_label)
    data = st.text_input('', value=st.session_state.data)

    if not data:
        st.stop()

    if is_plain:
        if st.session_state.regenerate or not st.session_state.columns or data != st.session_state.data:
            columns = get_noisy_columns(data, min_noise, max_noise)
            st.session_state.columns = columns
            unset_regenerate()
        else:
            columns = st.session_state.columns
    else:
        columns = data_to_columns(data, separator=' ')

    st.session_state.data = data

    st.subheader('\nColumns')
    st.text(visualize_columns(columns, delimiter=''))
    st.subheader('\n')

    placeholder = st.empty()
    recover_field, regen_filed = placeholder.columns([.085, 1])

    if is_plain:
        regen_filed.button('Regenerate', on_click=set_regenerate)

    if st.session_state.stop:
        stop_prediction()

    if columns and recover_field.button('Recover'):
        if st.session_state.is_unix:
            with placeholder.container():
                progress_bar_placeholder = st.empty()
                st.button('Stop', on_click=set_stop)

                with progress_bar_placeholder:
                    chains = predict(columns, bw)
        else:
            with placeholder:
                chains = predict(columns, bw)

        with placeholder.container():
            st.subheader('\nPrediction')
            st.text('\n\n'.join(chain for chain, _ in chains))

            if st.button('Clear'):
                st.session_state.task_id = None

        save_to_history(is_plain, min_noise, max_noise, bw, columns, chains)


def history_page() -> None:
    st.header('Inference History')

    if len(st.session_state.history) == 0:
        st.info('No records saved')
        return

    for record_id, (is_plain, text, min_noise, max_noise, bw, columns, chains) in enumerate(st.session_state.history,
                                                                                            start=1):
        st.info(f'Record {record_id}')

        if is_plain:
            st.text(f'Plain data: {text}')

        st.text(f'Noise range: [{min_noise}, {max_noise}]')
        st.text(f'Beam search width: {bw}')

        st.text('Columns:')
        st.text(visualize_columns(columns, delimiter=''))

        st.text('Prediction:')
        st.text('\n\n'.join(chain for chain, _ in chains))


if __name__ == '__main__':
    main()



import re
from pathlib import Path
from typing import List, Union

import numpy as np

from trecover.config import var


# ---------------------------------Plain inference data cleaning & preparation utils------------------------------------


def create_noisy_columns(data: str, min_noise: int, max_noise: int) -> List[str]:
    """
    Generate columns for keyless reading from plain data with defined noise range.

    Parameters
    ----------
    data : str
        Plain text.
    min_noise : int
        Minimum noise range value.
    max_noise : int
        Maximum noise range value.

    Returns
    -------
    columns : List[str]
        Columns for keyless reading.

    """

    np.random.seed(None)
    columns = list()

    data = re.sub(r'[^A-Za-z]', '', data).lower()

    for symbol in data:
        noise_size = np.random.randint(low=min_noise, high=max_noise, size=1)[0]
        noise_indexes = np.random.choice(list(var.ALPHABET.difference(symbol)), size=noise_size, replace=False)
        columns.append(f"{symbol}{''.join(noise_indexes)}")

    return columns


def create_files_noisy_columns(files: List[Union[str, Path]],
                               min_noise: int,
                               max_noise: int,
                               n_to_show: int = 0
                               ) -> List[List[str]]:
    """
    Generate columns for keyless reading from plain data contained in the files with defined noise range.

    Parameters
    ----------
    files : List[Union[str, Path]]
        Paths to files that contain plain data to create noised columns for keyless reading.
    min_noise : int
        Minimum noise range value.
    max_noise : int
        Maximum noise range value.
    n_to_show : int, default=0
        Maximum number of columns. Zero means no restrictions.

    Returns
    -------
    files_columns : List[List[str]]
        Batch of columns for keyless reading.

    """

    files_columns = list()

    for file in files:
        with open(file) as f:
            data = f.read()

        if n_to_show > 0:
            data = data[:n_to_show]

        columns = create_noisy_columns(data, min_noise, max_noise)

        files_columns.append(columns)

    return files_columns


# --------------------------------Noised inference data cleaning & preparation utils-----------------------------------


def data_to_columns(data: str, separator: str = ' ') -> List[str]:
    """
    Clean and split noised data.

    Parameters
    ----------
    data : str
        Noised columns for keyless reading.
    separator : str, default=' '
        Separator to split the data into columns.

    Returns
    -------
    List[str]:
        Columns for keyless reading.

    """

    data = re.sub(separator, ' ', data)
    cleaned_data = re.sub(r'[^A-Za-z ]', '', data).lower()

    return cleaned_data.split(' ')


def read_files_columns(files: List[Union[str, Path]], separator: str, n_to_show: int = 0) -> List[List[str]]:
    """
    Read, clean and split noised data contained in the files.

    Parameters
    ----------
    files : List[Union[str, Path]]
        Paths to files that contain noised data for keyless reading.
    separator : str
        Separator to split the data into columns.
    n_to_show : int, default=0
        Maximum number of columns. Zero means no restrictions.

    Returns
    -------
    files_columns : List[List[str]]
        Batch of columns for keyless reading.

    """

    files_columns = list()

    for file in files:
        with open(file) as f:
            data = f.read()

        columns = data_to_columns(data, separator)

        if n_to_show > 0:
            columns = columns[:n_to_show]

        files_columns.append(columns)

    return files_columns



from typing import List

import torch
from torch import Tensor

from trecover.config import var


def columns_to_tensor(columns: List[str], device: torch.device = torch.device('cpu')) -> Tensor:
    """
    Convert the columns to a torch tensor.

    Parameters
    ----------
    columns : List[str]
        Columns for keyless reading.
    device : torch.device, default=torch.device('cpu')
        The desired device of returned tensor.

    Returns
    -------
    tensor : Tensor[SEQUENCE_LEN, len(var.ALPHABET)]
        Columns as a torch tensor.

    """

    tensor = torch.zeros((len(columns), len(var.ALPHABET)), dtype=torch.float, device=device)

    for col in range(len(columns)):
        for symbol in columns[col]:
            tensor[col, var.ALPHABET2NUM[symbol]] = 1

    return tensor


def files_columns_to_tensors(files_columns: List[List[str]],
                             device: torch.device = torch.device('cpu')
                             ) -> List[Tensor]:
    """
    Convert the batch of columns to torch tensors.

    Parameters
    ----------
    files_columns : List[List[str]]
        Batch of columns for keyless reading.
    device : torch.device, default=torch.device('cpu')
        The desired device of returned tensor.

    Returns
    -------
    List[Tensor]:
        Columns batch as a list of torch tensors.

    """

    return [columns_to_tensor(columns, device) for columns in files_columns]


def tensor_to_columns(grid: Tensor) -> List[str]:
    """
    Convert the columns' tensor representation to a list of strings with alphabet symbols.

    Parameters
    ----------
    grid : Tensor[SEQUENCE_LEN, len(var.ALPHABET)]
        Columns for keyless reading as a tensor.

    Returns
    -------
    List[str]:
        Columns' tensor representation as a list of strings with alphabet symbols.

    """

    return [
        ''.join([var.NUM2ALPHABET[pos] for pos in range(grid.shape[1]) if grid[c, pos]])
        for c in range(grid.shape[0])
    ]


def tensor_to_target(tgt: Tensor) -> List[str]:
    """
    Convert the target tensor representation to list of alphabet symbols.

    Parameters
    ----------
    tgt : Tensor[SEQUENCE_LEN]
        Target tensor representation of columns' correct symbols.

    Returns
    -------
    List[str]:
        Target tensor representation to list of alphabet symbols.

    """

    return [var.NUM2ALPHABET[ch_id] for ch_id in tgt.tolist()]



from datetime import datetime
from functools import wraps
from http import HTTPStatus
from typing import Dict, Callable

from celery.result import AsyncResult
from fastapi import FastAPI, Request, Path

from trecover.app.api.backend.celeryapp import celery_app
from trecover.app.api.backend.tasks import predict, get_model_config
from trecover.app.api.schemas import PredictPayload, PredictResponse, TaskResponse
from trecover.config import log

api = FastAPI(title='TRecoverAPI', description='API for TRecover model')


@api.on_event('startup')
def startup() -> None:
    """ API startup handler. """

    log.project_logger.info('FatAPI launched')


def construct_response(handler: Callable[..., Dict]) -> Callable[..., Dict]:
    """
    A decorator that wraps a request handler.

    Parameters
    ----------
    handler : Callable[..., Dict]
        Request processing function.

    Returns
    -------
    wrap : Callable[..., Dict]
        Decorated handler.

    """

    @wraps(handler)
    def wrap(request: Request, *args, **kwargs) -> Dict:
        """
        A wrapper that constructs a JSON response for an endpoint's results.

        Parameters
        ----------
        request : Request
            Client request information.

        Returns
        -------
        response : Dict
            The result of the handler function.

        """

        response = handler(request, *args, **kwargs)

        response['method'] = request.method
        response['timestamp'] = datetime.now().isoformat()
        response['url'] = request.url._url

        return response

    return wrap


@api.get('/', tags=['General'])
@construct_response
def index(request: Request) -> Dict:
    """
    Healthcheck handler.

    Parameters
    ----------
    request : Request
        Client request information.

    Returns
    -------
    Dict:
        OK phrase as a Dict response.

    """

    return {
        'message': HTTPStatus.OK.phrase,
        'status_code': HTTPStatus.OK
    }


@api.get('/config', tags=['Configuration'])
@construct_response
def config(request: Request) -> Dict:
    """
    Get the configuration of the model used for inference.

    Parameters
    ----------
    request : Request
        Client request information.

    Returns
    -------
    response : Dict
        Response containing the values of the model configuration in the 'config' field.

    """

    task = get_model_config.delay()
    model_config = task.get()

    response = {
        'message': HTTPStatus.OK.phrase,
        'status_code': HTTPStatus.OK,
        'config': model_config
    }

    return response


@api.get('/config/{param}', tags=['Configuration'])
@construct_response
def config_param(request: Request, param: str) -> Dict:
    """
    Get the specific configuration parameter of the model used for inference.

    Parameters
    ----------
    request : Request
        Client request information.
    param : str
        Parameter name.

    Returns
    -------
    response : Dict
        Response containing the value of the specific configuration parameter in
        the '<param>' field if it exists, otherwise 'Not found' value.

    """

    task = get_model_config.delay()
    model_config = task.get()

    response = {
        'message': HTTPStatus.OK.phrase,
        'status_code': HTTPStatus.OK,
        param: model_config.get(param, 'Not found')
    }

    return response


@api.post('/recover', tags=['Prediction'], response_model=TaskResponse)
@construct_response
def recover(request: Request, payload: PredictPayload) -> Dict:
    """
    Perform keyless reading.

    Parameters
    ----------
    request : Request
        Client request information.
    payload : PredictPayload
        Data for keyless reading.

    Returns
    -------
    response : TaskResponse
        Response containing the id of the celery task in the 'task_id' field.

    """

    task = predict.delay(payload.columns, payload.beam_width, payload.delimiter)

    response = {
        'message': HTTPStatus.ACCEPTED.phrase,
        'status_code': HTTPStatus.ACCEPTED,
        'task_id': task.id
    }

    return response


@api.get('/status/{task_id}', tags=['Prediction'], response_model=PredictResponse)
@construct_response
def status(request: Request,
           task_id: str = Path(...,
                               title='The ID of the task to get status',
                               regex=r'[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}')
           ) -> Dict:
    """
    Get a celery task status.

    Parameters
    ----------
    request : Request
        Client request information.
    task_id : str
        Celery task id.

    Returns
    -------
    response : PredictResponse
        Response containing the status of the celery task in the 'state' and 'progress'
        fields if it's still in process,
        error information in 'message' and 'status_code' fields if it's failed,
        otherwise the result of keyless reading in the 'chains' field.

    """

    task = AsyncResult(task_id, app=celery_app)

    if task.failed():
        response = {
            'message': str(task.info),
            'status_code': HTTPStatus.CONFLICT
        }

    elif task.ready():
        columns, chains = task.get()

        response = {
            'message': HTTPStatus.OK.phrase,
            'status_code': HTTPStatus.OK,
            'columns': columns,
            'chains': chains,
            'progress': len(columns)
        }
    else:
        info = task.info

        response = {
            'message': HTTPStatus.PROCESSING.phrase,
            'status_code': HTTPStatus.PROCESSING,
            'state': task.status,
            'progress': info.get('progress') if isinstance(info, dict) else None
        }

    return response


@api.delete('/{task_id}', tags=['Prediction'])
@construct_response
def delete_prediction(request: Request,
                      task_id: str = Path(...,
                                          title='The ID of the task to forget',
                                          regex=r'[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}')
                      ) -> Dict:
    """
    Delete task result in the celery backend database.

    Parameters
    ----------
    request : Request
        Client request information.
    task_id : str
        Task ID to delete its result from celery backend database.

    Returns
    -------
    response : Dict
        OK phrase.

    """

    task = AsyncResult(task_id, app=celery_app)

    if task.ready():
        task.forget()
    else:
        task.revoke(terminate=True)

    response = {
        'message': HTTPStatus.OK.phrase,
        'status_code': HTTPStatus.OK
    }

    return response



from typing import Optional, List, Tuple

from pydantic import BaseModel, validator

from trecover.config import var


class PredictPayload(BaseModel):
    """ Content declaration of the request body to keyless read. """

    columns: List[str]
    beam_width: int
    delimiter: Optional[str] = ''

    @validator('columns')
    def columns_validator(cls, columns: List[str]) -> List[str]:
        """
        Check columns to keyless read.

        Parameters
        ----------
        columns: List[str]
            Columns to keyless read

        Returns
        -------
        data : List[str]
            Columns to keyless read.

        Raises
        ------
        ValueError:
            If at least one column is empty.

        """

        if any([len(column) == 0 for column in columns]):
            raise ValueError(f'Columns to keyless read must contain at least one character')

        return columns

    @validator('beam_width')
    def beam_width_validator(cls, beam_width: int) -> int:
        """
        Check columns to keyless read.

        Parameters
        ----------
        beam_width: int
            Columns to keyless read

        Returns
        -------
        beam_width : int
            Width of beam search algorithm.

        Raises
        ------
        ValueError:
            If the beam width is out of range.

        """

        if not 1 <= beam_width <= len(var.ALPHABET):
            raise ValueError(f'Beam width must be in between 1 and {len(var.ALPHABET)}')

        return beam_width

    class Config:
        """ PredictPayload example for API documentation"""

        schema_extra = {
            'example': {
                'columns': [
                    'adin', 'scjz', 'pzxz', 'evft', 'odev', 'pkzc', 'lipss', 'eayh', 'aepqc', 'rvuqh', 'ozldx',
                    'ulgq', 'nabhr', 'dvvf', 'tmil', 'huiow', 'euzq', 'cdmh', 'ompzv', 'uguq', 'ntvqi', 'tctm',
                    'rwgi', 'yuonh', 'wkvr', 'efau', 'nrele', 'tbhxq', 'iiqd', 'nhwfy', 'tymio', 'onzpj', 'turho',
                    'hhgnd', 'ehmof', 'scpoi', 'tyqbn', 'roul', 'exoe', 'edwyc', 'thid', 'snqrp', 'tmmh', 'oejg',
                    'cjbaw', 'hgzmb', 'ezlwj', 'esxb', 'rria', 'taeo', 'hpnln', 'eeomf', 'cvxr', 'ofwy', 'nuon',
                    'vvex', 'ikbd', 'cmusb', 'tdnkb', 'iccd', 'ocay', 'nnxds', 'slbgd', 'oupap', 'mhip', 'ebtv',
                    'bipta', 'unazv', 'sccol', 'iovq', 'nxzev', 'ezaz', 'sakm', 'skls', 'eipvd', 'segz', 'inrlo',
                    'nzkug', 'puau', 'oolt', 'recs', 'tjot', 'lwds', 'assb', 'ntyzf', 'dpad', 'byep', 'owgol',
                    'aiwbb', 'ralk', 'dffos', 'ejfp', 'dwyug', 'uxucm', 'ppch', 'teznz', 'hpfbv', 'eaocf', 'iist',
                    'rdlyx', 'wseu', 'ivgg', 'nabi', 'dtuiw', 'oxzje', 'wderj', 'sawm', 'odpw', 'noqs', 'cimnf',
                    'eukjx', 'aszy', 'ghmiq', 'ajcsd', 'ipwgf', 'nhth', 'tpvfp', 'hnede', 'aclb', 'tteuq', 'nbvi',
                    'isgc', 'gejbz', 'hufnl', 'tggvn', 'aahfn', 'skkxl', 'mmwq', 'anqgw', 'lrvji', 'lker', 'gvqt',
                    'rdcd', 'obvt', 'ussht', 'piwb', 'ozbn', 'fyig', 'aczx', 'clak', 'tgnxi', 'iieyi', 'vqyqo',
                    'ijgp', 'shfth', 'tnbeg', 'sbuzx', 'wwnja', 'eguhn', 'aneai', 'rpsl', 'izut', 'njixa', 'grgh',
                    'bkypw', 'lliv', 'azeob', 'cojl', 'kmqnd', 'ahpq', 'pckw', 'phvpa', 'rgqv', 'opqby', 'aiqsp',
                    'clirf', 'hlrp', 'eztn', 'dgek', 'azubn', 'gnskd', 'rajo', 'okgv', 'ugcp', 'pilf', 'oxbx',
                    'fwqq', 'jsmf', 'okgw', 'uvjyg', 'rfio', 'nbtcy', 'aryp', 'ledgj', 'inex', 'sdqmm', 'tcmtw',
                    'soqh', 'tquu', 'hrfi', 'rsddp', 'ewfye', 'aumxj', 'tssui', 'ecgd', 'nunfp', 'isfka', 'ncvx',
                    'gqxi', 'typc', 'okcdt', 'sgcaq', 'mbcwd', 'angb', 'svyl', 'hjmk', 'tlcgd', 'hpka', 'efule',
                    'cbjou', 'agjjr', 'mziur', 'emoqh', 'rajte', 'ahoz', 'shaxe', 'okeyj', 'fgwnq', 'twojw',
                    'hcyfc', 'ojehd', 'soij', 'ekhpc', 'wbyq', 'hdfi', 'opyas', 'rzba', 'ediml', 'mvwys', 'avtmz',
                    'ilvj', 'nefy', 'eycng', 'dqdk', 'oihcj', 'nzuw', 'srrye', 'ccsi', 'emhnz', 'neosn', 'ewzki'
                ],
                'beam_width': 5,
                'delimiter': ''
            }
        }


class BaseResponse(BaseModel):
    """ Basic contents declaration of the response body for all API handlers. """

    message: str
    method: str
    status_code: int
    timestamp: str
    url: str


class TaskResponse(BaseResponse):
    """ Contents declaration of the 'recover' handler response body. """

    task_id: Optional[str]

    class Config:
        """ TaskResponse example for API documentation"""

        schema_extra = {
            'example': {
                "message": "Accepted",
                "method": "POST",
                "status_code": 202,
                "timestamp": "2021-09-04T12:53:35.512412",
                "url": "http://localhost:8001/recover",
                "task_id": "909e4817-cca3-4dbf-a598-f7f83c5d60c9"
            }
        }


class PredictResponse(BaseResponse):
    """ Contents declaration of the 'status' handler response body. """

    columns: Optional[List[str]]
    chains: Optional[List[Tuple]]
    state: Optional[str]
    progress: Optional[int]

    class Config:
        """ PredictResponse example for API documentation"""

        schema_extra = {
            'example': {
                "message": "OK",
                "method": "GET",
                "status_code": 200,
                "timestamp": "2021-09-04T12:54:49.801614",
                "url": "http://localhost:8001/status/909e4817-cca3-4dbf-a598-f7f83c5d60c9",
                "columns": [
                    "adin", "scjz", "pzxz", "evft", "odev", "pkzc", "lipss", "eayh", "aepqc", "rvuqh", "ozldx", "ulgq",
                    "nabhr", "dvvf", "tmil", "huiow", "euzq", "cdmh", "ompzv", "uguq", "ntvqi", "tctm", "rwgi", "yuonh",
                    "wkvr", "efau", "nrele", "tbhxq", "iiqd", "nhwfy", "tymio", "onzpj", "turho", "hhgnd", "ehmof",
                    "scpoi", "tyqbn", "roul", "exoe", "edwyc", "thid", "snqrp", "tmmh", "oejg", "cjbaw", "hgzmb",
                    "ezlwj", "esxb", "rria", "taeo", "hpnln", "eeomf", "cvxr", "ofwy", "nuon", "vvex", "ikbd", "cmusb",
                    "tdnkb", "iccd", "ocay", "nnxds", "slbgd", "oupap", "mhip", "ebtv", "bipta", "unazv", "sccol",
                    "iovq", "nxzev", "ezaz", "sakm", "skls", "eipvd", "segz", "inrlo", "nzkug", "puau", "oolt", "recs",
                    "tjot", "lwds", "assb", "ntyzf", "dpad", "byep", "owgol", "aiwbb", "ralk", "dffos", "ejfp", "dwyug",
                    "uxucm", "ppch", "teznz", "hpfbv", "eaocf", "iist", "rdlyx", "wseu", "ivgg", "nabi", "dtuiw",
                    "oxzje", "wderj", "sawm", "odpw", "noqs", "cimnf", "eukjx", "aszy", "ghmiq", "ajcsd", "ipwgf",
                    "nhth", "tpvfp", "hnede", "aclb", "tteuq", "nbvi", "isgc", "gejbz", "hufnl", "tggvn", "aahfn",
                    "skkxl", "mmwq", "anqgw", "lrvji", "lker", "gvqt", "rdcd", "obvt", "ussht", "piwb", "ozbn", "fyig",
                    "aczx", "clak", "tgnxi", "iieyi", "vqyqo", "ijgp", "shfth", "tnbeg", "sbuzx", "wwnja", "eguhn",
                    "aneai", "rpsl", "izut", "njixa", "grgh", "bkypw", "lliv", "azeob", "cojl", "kmqnd", "ahpq", "pckw",
                    "phvpa", "rgqv", "opqby", "aiqsp", "clirf", "hlrp", "eztn", "dgek", "azubn", "gnskd", "rajo",
                    "okgv", "ugcp", "pilf", "oxbx", "fwqq", "jsmf", "okgw", "uvjyg", "rfio", "nbtcy", "aryp", "ledgj",
                    "inex", "sdqmm", "tcmtw", "soqh", "tquu", "hrfi", "rsddp", "ewfye", "aumxj", "tssui", "ecgd",
                    "nunfp", "isfka", "ncvx", "gqxi", "typc", "okcdt", "sgcaq", "mbcwd", "angb", "svyl", "hjmk",
                    "tlcgd", "hpka", "efule", "cbjou", "agjjr", "mziur", "emoqh", "rajte", "ahoz", "shaxe", "okeyj",
                    "fgwnq", "twojw", "hcyfc", "ojehd", "soij", "ekhpc", "wbyq", "hdfi", "opyas", "rzba", "ediml",
                    "mvwys", "avtmz", "ilvj", "nefy", "eycng", "dqdk", "oihcj", "nzuw", "srrye", "ccsi", "emhnz",
                    "neosn", "ewzki"

                ],
                "chains": [
                    [
                        "aspeoplearoundthecountrywentintothestreetstoabeertheconvictionsametalineaspeonportsandboarded"
                        "uptheirwindowsonceagainthatnightasmallgroupofactivistswearingbloodapproachedagroupoffourtales"
                        "courseateningtoawalkthecametheonthesewisremainedourcesw",
                        -33.01948561671179
                    ],
                    [
                        "aspeoplearoundthecountrywentintothestreetstoabeertheconvictionsametalineaspeonportsandboarded"
                        "uptheirwindowsdoneagainthatnightasmallgroupofactivistswearingbloodapproachedagroupoffourtales"
                        "courseateningtoawalkthecametheonthesewisremainedourcesw",
                        -33.05044434355386
                    ],
                    [
                        "aspeoplearoundthecountrywentintothestreetstoabeertheconvictionsametalineaspeonportsandboarded"
                        "uptheirwindowsdoneagainthatnightasmallgroupofactivistswearingbloodapproachedagroupoffourtales"
                        "courseateningtoawalkthecametheonthesewisremainedourchew",
                        -33.686079347001396
                    ],
                    [
                        "aspeoplearoundthecountrywentintothestreetstoabeertheconvictionsametalineaspeonportsandboarded"
                        "uptheirwindowsdoneagainthatnightasmallgroupofactivistswearingbloodapproachedagroupoffourtales"
                        "coursexteningtoawalkthecametheonthesewisremainedourchew",
                        -33.83420559252136
                    ],
                    [
                        "aspeoplearoundthecountrywentintothestreetstoabeertheconvictionsametalineaspeonportsandboarded"
                        "uptheirwindowsonceagainthatnightasmallgroupofactivistswearingbloodapproachedagroupoffourtales"
                        "courseateningtoawalkthecametheonthesewisremainedourchew",
                        -33.948646204201395
                    ]
                ],
                "state": None,
                "progress": None
            }
        }
